# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class ExtensionDriver(object):
    """Define stable abstract interface for extension drivers.
    An extension driver extends the core resources implemented by the
    plugin with additional attributes. Methods that process create
    and update operations for these resources validate and persist
    values for extended attributes supplied through the API. Other
    methods extend the resource dictionaries returned from the API
    operations with the values of the extended attributes.
    """

    @abc.abstractmethod
    def initialize(self):
        """Perform driver initialization.
        Called after all drivers have been loaded and the database has
        been initialized. No abstract methods defined below will be
        called prior to this method being called.
        """
        pass

    @property
    def extension_alias(self):
        """Supported extension alias.
        Return the alias identifying the core API extension supported
        by this driver. Do not declare if API extension handling will
        be left to a service plugin, and we just need to provide
        core resource extension and updates.
        """
        pass

    def process_create_network(self, plugin_context, data, result):
        """Process extended attributes for create network.
        :param plugin_context: plugin request context
        :param data: dictionary of incoming network data
        :param result: network dictionary to extend
        Called inside transaction context on plugin_context.session to
        validate and persist any extended network attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_create_subnet(self, plugin_context, data, result):
        """Process extended attributes for create subnet.
        :param plugin_context: plugin request context
        :param data: dictionary of incoming subnet data
        :param result: subnet dictionary to extend
        Called inside transaction context on plugin_context.session to
        validate and persist any extended subnet attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_create_port(self, plugin_context, data, result):
        """Process extended attributes for create port.
        :param plugin_context: plugin request context
        :param data: dictionary of incoming port data
        :param result: port dictionary to extend
        Called inside transaction context on plugin_context.session to
        validate and persist any extended port attributes defined by this
        driver. Extended attribute values must also be added to
        result.
        """
        pass

    def process_update_network(self, plugin_context, data, result):
        """Process extended attributes for update network.
        :param plugin_context: plugin request context
        :param data: dictionary of incoming network data
        :param result: network dictionary to extend
        Called inside transaction context on plugin_context.session to
        validate and update any extended network attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def process_update_subnet(self, plugin_context, data, result):
        """Process extended attributes for update subnet.
        :param plugin_context: plugin request context
        :param data: dictionary of incoming subnet data
        :param result: subnet dictionary to extend
        Called inside transaction context on plugin_context.session to
        validate and update any extended subnet attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def process_update_port(self, plugin_context, data, result):
        """Process extended attributes for update port.
        :param plugin_context: plugin request context
        :param data: dictionary of incoming port data
        :param result: port dictionary to extend
        Called inside transaction context on plugin_context.session to
        validate and update any extended port attributes defined by this
        driver. Extended attribute values, whether updated or not,
        must also be added to result.
        """
        pass

    def extend_network_dict(self, session, base_model, result):
        """Add extended attributes to network dictionary.
        :param session: database session
        :param base_model: network model data
        :param result: network dictionary to extend
        Called inside transaction context on session to add any
        extended attributes defined by this driver to a network
        dictionary to be used for driver calls and/or
        returned as the result of a network operation.
        """
        pass

    def extend_subnet_dict(self, session, base_model, result):
        """Add extended attributes to subnet dictionary.
        :param session: database session
        :param base_model: subnet model data
        :param result: subnet dictionary to extend
        Called inside transaction context on session to add any
        extended attributes defined by this driver to a subnet
        dictionary to be used for driver calls and/or
        returned as the result of a subnet operation.
        """
        pass

    def extend_port_dict(self, session, base_model, result):
        """Add extended attributes to port dictionary.
        :param session: database session
        :param base_model: port model data
        :param result: port dictionary to extend
        Called inside transaction context on session to add any
        extended attributes defined by this driver to a port
        dictionary to be used for driver calls
        and/or returned as the result of a port operation.
        """
        pass



# Copyright 2013 VMware, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron.tests import base
from neutron_lib import exceptions
from oslo_serialization import jsonutils
import six

from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import utils
from vmware_nsx.nsxlib.mh import lsn as lsnlib


class LSNTestCase(base.BaseTestCase):

    def setUp(self):
        super(LSNTestCase, self).setUp()
        self.mock_request_p = mock.patch(
            'vmware_nsx.nsxlib.mh.do_request')
        self.mock_request = self.mock_request_p.start()
        self.cluster = mock.Mock()
        self.cluster.default_service_cluster_uuid = 'foo'

    def test_service_cluster_None(self):
        self.mock_request.return_value = None
        expected = lsnlib.service_cluster_exists(None, None)
        self.assertFalse(expected)

    def test_service_cluster_found(self):
        self.mock_request.return_value = {
            "results": [
                {
                    "_href": "/ws.v1/service-cluster/foo_uuid",
                    "display_name": "foo_name",
                    "uuid": "foo_uuid",
                    "tags": [],
                    "_schema": "/ws.v1/schema/ServiceClusterConfig",
                    "gateways": []
                }
            ],
            "result_count": 1
        }
        expected = lsnlib.service_cluster_exists(None, 'foo_uuid')
        self.assertTrue(expected)

    def test_service_cluster_not_found(self):
        self.mock_request.side_effect = exceptions.NotFound()
        expected = lsnlib.service_cluster_exists(None, 'foo_uuid')
        self.assertFalse(expected)

    def test_lsn_for_network_create(self):
        net_id = "foo_network_id"
        tags = utils.get_tags(n_network_id=net_id)
        obj = {"edge_cluster_uuid": "foo", "tags": tags}
        lsnlib.lsn_for_network_create(self.cluster, net_id)
        self.mock_request.assert_called_once_with(
            "POST", "/ws.v1/lservices-node",
            jsonutils.dumps(obj, sort_keys=True), cluster=self.cluster)

    def test_lsn_for_network_get(self):
        net_id = "foo_network_id"
        lsn_id = "foo_lsn_id"
        self.mock_request.return_value = {
            "results": [{"uuid": "foo_lsn_id"}],
            "result_count": 1
        }
        result = lsnlib.lsn_for_network_get(self.cluster, net_id)
        self.assertEqual(lsn_id, result)
        self.mock_request.assert_called_once_with(
            "GET",
            ("/ws.v1/lservices-node?fields=uuid&tag=%s&"
             "tag_scope=n_network_id" % net_id),
            cluster=self.cluster)

    def test_lsn_for_network_get_none(self):
        net_id = "foo_network_id"
        self.mock_request.return_value = {
            "results": [{"uuid": "foo_lsn_id1"}, {"uuid": "foo_lsn_id2"}],
            "result_count": 2
        }
        result = lsnlib.lsn_for_network_get(self.cluster, net_id)
        self.assertIsNone(result)

    def test_lsn_for_network_get_raise_not_found(self):
        net_id = "foo_network_id"
        self.mock_request.return_value = {
            "results": [], "result_count": 0
        }
        self.assertRaises(exceptions.NotFound,
                          lsnlib.lsn_for_network_get,
                          self.cluster, net_id)

    def test_lsn_delete(self):
        lsn_id = "foo_id"
        lsnlib.lsn_delete(self.cluster, lsn_id)
        self.mock_request.assert_called_once_with(
            "DELETE",
            "/ws.v1/lservices-node/%s" % lsn_id, cluster=self.cluster)

    def _test_lsn_port_host_entries_update(self, lsn_type, hosts_data):
        lsn_id = 'foo_lsn_id'
        lsn_port_id = 'foo_lsn_port_id'
        lsnlib.lsn_port_host_entries_update(
            self.cluster, lsn_id, lsn_port_id, lsn_type, hosts_data)
        self.mock_request.assert_called_once_with(
            'PUT',
            '/ws.v1/lservices-node/%s/lport/%s/%s' % (lsn_id,
                                                      lsn_port_id,
                                                      lsn_type),
            jsonutils.dumps({'hosts': hosts_data}, sort_keys=True),
            cluster=self.cluster)

    def test_lsn_port_dhcp_entries_update(self):
        hosts_data = [{"ip_address": "11.22.33.44",
                       "mac_address": "aa:bb:cc:dd:ee:ff"},
                      {"ip_address": "44.33.22.11",
                       "mac_address": "ff:ee:dd:cc:bb:aa"}]
        self._test_lsn_port_host_entries_update("dhcp", hosts_data)

    def test_lsn_port_metadata_entries_update(self):
        hosts_data = [{"ip_address": "11.22.33.44",
                       "device_id": "foo_vm_uuid"}]
        self._test_lsn_port_host_entries_update("metadata-proxy", hosts_data)

    def test_lsn_port_create(self):
        port_data = {
            "ip_address": "1.2.3.0/24",
            "mac_address": "aa:bb:cc:dd:ee:ff",
            "subnet_id": "foo_subnet_id"
        }
        port_id = "foo_port_id"
        self.mock_request.return_value = {"uuid": port_id}
        lsn_id = "foo_lsn_id"
        result = lsnlib.lsn_port_create(self.cluster, lsn_id, port_data)
        self.assertEqual(result, port_id)
        tags = utils.get_tags(n_subnet_id=port_data["subnet_id"],
                              n_mac_address=port_data["mac_address"])
        port_obj = {
            "ip_address": port_data["ip_address"],
            "mac_address": port_data["mac_address"],
            "type": "LogicalServicesNodePortConfig",
            "tags": tags
        }
        self.mock_request.assert_called_once_with(
            "POST", "/ws.v1/lservices-node/%s/lport" % lsn_id,
            jsonutils.dumps(port_obj, sort_keys=True), cluster=self.cluster)

    def test_lsn_port_delete(self):
        lsn_id = "foo_lsn_id"
        lsn_port_id = "foo_port_id"
        lsnlib.lsn_port_delete(self.cluster, lsn_id, lsn_port_id)
        self.mock_request.assert_called_once_with(
            "DELETE",
            "/ws.v1/lservices-node/%s/lport/%s" % (lsn_id, lsn_port_id),
            cluster=self.cluster)

    def test_lsn_port_get_with_filters(self):
        lsn_id = "foo_lsn_id"
        port_id = "foo_port_id"
        filters = {"tag": "foo_tag", "tag_scope": "foo_scope"}
        self.mock_request.return_value = {
            "results": [{"uuid": port_id}],
            "result_count": 1
        }
        result = lsnlib._lsn_port_get(self.cluster, lsn_id, filters)
        self.assertEqual(result, port_id)
        self.mock_request.assert_called_once_with(
            "GET",
            ("/ws.v1/lservices-node/%s/lport?fields=uuid&tag=%s&"
             "tag_scope=%s" % (lsn_id, filters["tag"], filters["tag_scope"])),
            cluster=self.cluster)

    def test_lsn_port_get_with_filters_return_none(self):
        self.mock_request.return_value = {
            "results": [{"uuid": "foo1"}, {"uuid": "foo2"}],
            "result_count": 2
        }
        result = lsnlib._lsn_port_get(self.cluster, "lsn_id", None)
        self.assertIsNone(result)

    def test_lsn_port_get_with_filters_raises_not_found(self):
        self.mock_request.return_value = {"results": [], "result_count": 0}
        self.assertRaises(exceptions.NotFound,
                          lsnlib._lsn_port_get,
                          self.cluster, "lsn_id", None)

    def test_lsn_port_info_get(self):
        self.mock_request.return_value = {
            "tags": [
                {"scope": "n_mac_address", "tag": "fa:16:3e:27:fd:a0"},
                {"scope": "n_subnet_id", "tag": "foo_subnet_id"},
            ],
            "mac_address": "aa:bb:cc:dd:ee:ff",
            "ip_address": "0.0.0.0/0",
            "uuid": "foo_lsn_port_id"
        }
        result = lsnlib.lsn_port_info_get(
            self.cluster, 'foo_lsn_id', 'foo_lsn_port_id')
        self.mock_request.assert_called_once_with(
            'GET', '/ws.v1/lservices-node/foo_lsn_id/lport/foo_lsn_port_id',
            cluster=self.cluster)
        self.assertIn('subnet_id', result)
        self.assertIn('mac_address', result)

    def test_lsn_port_info_get_raise_not_found(self):
        self.mock_request.side_effect = exceptions.NotFound
        self.assertRaises(exceptions.NotFound,
                          lsnlib.lsn_port_info_get,
                          self.cluster, mock.ANY, mock.ANY)

    def test_lsn_port_plug_network(self):
        lsn_id = "foo_lsn_id"
        lsn_port_id = "foo_lsn_port_id"
        lswitch_port_id = "foo_lswitch_port_id"
        lsnlib.lsn_port_plug_network(
            self.cluster, lsn_id, lsn_port_id, lswitch_port_id)
        self.mock_request.assert_called_once_with(
            "PUT",
            ("/ws.v1/lservices-node/%s/lport/%s/"
             "attachment") % (lsn_id, lsn_port_id),
            jsonutils.dumps({"peer_port_uuid": lswitch_port_id,
                             "type": "PatchAttachment"}, sort_keys=True),
            cluster=self.cluster)

    def test_lsn_port_plug_network_raise_conflict(self):
        lsn_id = "foo_lsn_id"
        lsn_port_id = "foo_lsn_port_id"
        lswitch_port_id = "foo_lswitch_port_id"
        self.mock_request.side_effect = api_exc.Conflict
        self.assertRaises(
            nsx_exc.LsnConfigurationConflict,
            lsnlib.lsn_port_plug_network,
            self.cluster, lsn_id, lsn_port_id, lswitch_port_id)

    def _test_lsn_port_dhcp_configure(
        self, lsn_id, lsn_port_id, is_enabled, opts):
        lsnlib.lsn_port_dhcp_configure(
            self.cluster, lsn_id, lsn_port_id, is_enabled, opts)
        opt_array = [
            {"name": key, "value": val}
            for key, val in six.iteritems(opts)
        ]
        self.mock_request.assert_has_calls([
            mock.call("PUT", "/ws.v1/lservices-node/%s/dhcp" % lsn_id,
                      jsonutils.dumps({"enabled": is_enabled},
                                      sort_keys=True),
                      cluster=self.cluster),
            mock.call("PUT",
                      ("/ws.v1/lservices-node/%s/"
                       "lport/%s/dhcp") % (lsn_id, lsn_port_id),
                      jsonutils.dumps({"options": opt_array}, sort_keys=True),
                      cluster=self.cluster)
        ])

    def test_lsn_port_dhcp_configure_empty_opts(self):
        lsn_id = "foo_lsn_id"
        lsn_port_id = "foo_lsn_port_id"
        is_enabled = False
        opts = {}
        self._test_lsn_port_dhcp_configure(
            lsn_id, lsn_port_id, is_enabled, opts)

    def test_lsn_port_dhcp_configure_with_opts(self):
        lsn_id = "foo_lsn_id"
        lsn_port_id = "foo_lsn_port_id"
        is_enabled = True
        opts = {"opt1": "val1", "opt2": "val2"}
        self._test_lsn_port_dhcp_configure(
            lsn_id, lsn_port_id, is_enabled, opts)

    def _test_lsn_metadata_configure(
        self, lsn_id, is_enabled, opts, expected_opts):
        lsnlib.lsn_metadata_configure(
            self.cluster, lsn_id, is_enabled, opts)
        lsn_obj = {"enabled": is_enabled}
        lsn_obj.update(expected_opts)
        self.mock_request.assert_has_calls([
            mock.call("PUT",
                      "/ws.v1/lservices-node/%s/metadata-proxy" % lsn_id,
                      jsonutils.dumps(lsn_obj, sort_keys=True),
                      cluster=self.cluster),
        ])

    def test_lsn_port_metadata_configure_empty_secret(self):
        lsn_id = "foo_lsn_id"
        is_enabled = True
        opts = {
            "metadata_server_ip": "1.2.3.4",
            "metadata_server_port": "8775"
        }
        expected_opts = {
            "metadata_server_ip": "1.2.3.4",
            "metadata_server_port": "8775",
        }
        self._test_lsn_metadata_configure(
            lsn_id, is_enabled, opts, expected_opts)

    def test_lsn_metadata_configure_with_secret(self):
        lsn_id = "foo_lsn_id"
        is_enabled = True
        opts = {
            "metadata_server_ip": "1.2.3.4",
            "metadata_server_port": "8775",
            "metadata_proxy_shared_secret": "foo_secret"
        }
        expected_opts = {
            "metadata_server_ip": "1.2.3.4",
            "metadata_server_port": "8775",
            "options": [{
                "name": "metadata_proxy_shared_secret",
                "value": "foo_secret"
            }]
        }
        self._test_lsn_metadata_configure(
            lsn_id, is_enabled, opts, expected_opts)

    def _test_lsn_port_host_action(
            self, lsn_port_action_func, extra_action, action, host):
        lsn_id = "foo_lsn_id"
        lsn_port_id = "foo_lsn_port_id"
        lsn_port_action_func(self.cluster, lsn_id, lsn_port_id, host)
        self.mock_request.assert_called_once_with(
            "POST",
            ("/ws.v1/lservices-node/%s/lport/"
             "%s/%s?action=%s") % (lsn_id, lsn_port_id, extra_action, action),
            jsonutils.dumps(host, sort_keys=True), cluster=self.cluster)

    def test_lsn_port_dhcp_host_add(self):
        host = {
            "ip_address": "1.2.3.4",
            "mac_address": "aa:bb:cc:dd:ee:ff"
        }
        self._test_lsn_port_host_action(
            lsnlib.lsn_port_dhcp_host_add, "dhcp", "add_host", host)

    def test_lsn_port_dhcp_host_remove(self):
        host = {
            "ip_address": "1.2.3.4",
            "mac_address": "aa:bb:cc:dd:ee:ff"
        }
        self._test_lsn_port_host_action(
            lsnlib.lsn_port_dhcp_host_remove, "dhcp", "remove_host", host)

    def test_lsn_port_metadata_host_add(self):
        host = {
            "ip_address": "1.2.3.4",
            "instance_id": "foo_instance_id"
        }
        self._test_lsn_port_host_action(lsnlib.lsn_port_metadata_host_add,
                                        "metadata-proxy", "add_host", host)

    def test_lsn_port_metadata_host_remove(self):
        host = {
            "ip_address": "1.2.3.4",
            "instance_id": "foo_instance_id"
        }
        self._test_lsn_port_host_action(lsnlib.lsn_port_metadata_host_remove,
                                        "metadata-proxy", "remove_host", host)



# Copyright 2015 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

import mock

from oslo_config import cfg
from oslo_log import log as logging

from neutron.db import common_db_mixin as common_db
from neutron_lib import context as neutron_context
from neutron_lib.plugins import directory

from vmware_nsx.common import config
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx import plugin
from vmware_nsx.plugins.nsx_v.vshield import vcns
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils

LOG = logging.getLogger(__name__)


def get_nsxv_client():
    return vcns.Vcns(
        address=cfg.CONF.nsxv.manager_uri,
        user=cfg.CONF.nsxv.user,
        password=cfg.CONF.nsxv.password,
        ca_file=cfg.CONF.nsxv.ca_file,
        insecure=cfg.CONF.nsxv.insecure)


def get_plugin_filters(context):
    return admin_utils.get_plugin_filters(
        context, projectpluginmap.NsxPlugins.NSX_V)


class NeutronDbClient(common_db.CommonDbMixin):
    def __init__(self):
        super(NeutronDbClient, self)
        self.context = neutron_context.get_admin_context()


class NsxVPluginWrapper(plugin.NsxVPlugin):

    def __init__(self):
        config.register_nsxv_azs(cfg.CONF, cfg.CONF.nsxv.availability_zones)
        self.context = neutron_context.get_admin_context()
        self.filters = get_plugin_filters(self.context)
        super(NsxVPluginWrapper, self).__init__()
        # Make this the core plugin
        directory.add_plugin('CORE', self)
        # finish the plugin initialization
        # (with md-proxy config, but without housekeeping)
        with mock.patch("vmware_nsx.plugins.common.housekeeper."
                        "housekeeper.NsxvHousekeeper"):
            self.init_complete(0, 0, 0)

    def _start_rpc_listeners(self):
        pass

    def _extend_get_network_dict_provider(self, context, net):
        self._extend_network_dict_provider(context, net)
        # skip getting the Qos policy ID because get_object calls
        # plugin init again on admin-util environment

    def count_spawn_jobs(self):
        # check if there are any spawn jobs running
        return self.edge_manager._get_worker_pool().running()

    # Define enter & exit to be used in with statements
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        """Wait until no more jobs are pending

        We want to wait until all spawn edge creation are done, or else the
        edges might be in PERNDING_CREATE state in the nsx DB
        """
        if not self.count_spawn_jobs():
            return

        LOG.warning("Waiting for plugin jobs to finish properly...")
        sleep_time = 1
        print_time = 20
        max_loop = 600
        for print_index in range(1, max_loop):
            n_jobs = self.count_spawn_jobs()
            if n_jobs > 0:
                if (print_index % print_time) == 0:
                    LOG.warning("Still Waiting on %(jobs)s "
                                "job%(plural)s",
                                {'jobs': n_jobs,
                                 'plural': 's' if n_jobs > 1 else ''})
                time.sleep(sleep_time)
            else:
                LOG.warning("Done.")
                return

        LOG.warning("Sorry. Waited for too long. Some jobs are still "
                    "running.")

    def _update_filters(self, requested_filters):
        filters = self.filters.copy()
        if requested_filters:
            filters.update(requested_filters)
        return filters

    def get_networks(self, context, filters=None, fields=None):
        filters = self._update_filters(filters)
        return super(NsxVPluginWrapper, self).get_networks(
            context, filters=filters, fields=fields)

    def get_subnets(self, context, filters=None, fields=None):
        filters = self._update_filters(filters)
        return super(NsxVPluginWrapper, self).get_subnets(
            context, filters=filters, fields=fields)

    def get_ports(self, context, filters=None, fields=None):
        filters = self._update_filters(filters)
        return super(NsxVPluginWrapper, self).get_ports(
            self.context, filters=filters, fields=fields)

    def get_routers(self, context, filters=None, fields=None):
        filters = self._update_filters(filters)
        return super(NsxVPluginWrapper, self).get_routers(
            self.context, filters=filters, fields=fields)


def get_nsxv_backend_edges():
    """Get a list of all the backend edges and some of their attributes
    """
    nsxv = get_nsxv_client()
    edges = nsxv.get_edges()
    backend_edges = []
    for edge in edges:
        summary = edge.get('appliancesSummary')
        size = ha = None
        if summary:
            size = summary.get('applianceSize')
            deployed_vms = summary.get('numberOfDeployedVms', 1)
            ha = 'Enabled' if deployed_vms > 1 else 'Disabled'
        # get all the relevant backend information for this edge
        edge_data = {
            'id': edge.get('id'),
            'name': edge.get('name'),
            'size': size,
            'type': edge.get('edgeType'),
            'ha': ha,
        }
        backend_edges.append(edge_data)
    return backend_edges


def get_edge_syslog_info(edge_id):
    """Get syslog information for specific edge id"""

    nsxv = get_nsxv_client()
    syslog_info = nsxv.get_edge_syslog(edge_id)[1]
    if not syslog_info['enabled']:
        return 'Disabled'

    output = ""
    if 'protocol' in syslog_info:
        output += syslog_info['protocol']
    if 'serverAddresses' in syslog_info:
        for server_address in syslog_info['serverAddresses']['ipAddress']:
            output += "\n" + server_address

    return output



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from neutron_lib.exceptions import firewall_v2 as exceptions
from neutron_lib.plugins import directory

from vmware_nsx.services.fwaas.nsx_v3 import edge_fwaas_driver_base
from vmware_nsx.services.fwaas.nsx_v3 import edge_fwaas_driver_v2
from vmware_nsx.services.fwaas.nsx_v3 import fwaas_callbacks_v2
from vmware_nsx.tests.unit.nsx_v3 import test_plugin as test_v3_plugin
from vmware_nsxlib.v3 import nsx_constants as consts

FAKE_FW_ID = 'fake_fw_uuid'
FAKE_ROUTER_ID = 'fake_rtr_uuid'
FAKE_PORT_ID = 'fake_port_uuid'
FAKE_NET_ID = 'fake_net_uuid'
FAKE_NSX_LS_ID = 'fake_nsx_ls_uuid'
MOCK_NSX_ID = 'nsx_nsx_router_id'
MOCK_DEFAULT_RULE_ID = 'nsx_default_rule_id'
MOCK_SECTION_ID = 'sec_id'
DEFAULT_RULE = {'is_default': True,
                'display_name': edge_fwaas_driver_base.DEFAULT_RULE_NAME,
                'id': MOCK_DEFAULT_RULE_ID,
                'action': consts.FW_ACTION_DROP}


class Nsxv3FwaasTestCase(test_v3_plugin.NsxV3PluginTestCaseMixin):
    def setUp(self):
        super(Nsxv3FwaasTestCase, self).setUp()
        self.firewall = edge_fwaas_driver_v2.EdgeFwaasV3DriverV2()

        # Start some nsxlib/DB mocks
        mock.patch(
            "vmware_nsxlib.v3.core_resources.NsxLibLogicalRouter."
            "get_firewall_section_id",
            return_value=MOCK_SECTION_ID).start()

        mock.patch(
            "vmware_nsxlib.v3.security.NsxLibFirewallSection."
            "get_default_rule",
            return_value={'id': MOCK_DEFAULT_RULE_ID}).start()

        mock.patch(
            "vmware_nsx.db.db.get_nsx_router_id",
            return_value=MOCK_NSX_ID).start()

        self.plugin = directory.get_plugin()
        self.plugin.fwaas_callbacks = fwaas_callbacks_v2.\
            Nsxv3FwaasCallbacksV2()
        self.plugin.fwaas_callbacks.fwaas_enabled = True
        self.plugin.fwaas_callbacks.fwaas_driver = self.firewall
        self.plugin.fwaas_callbacks.internal_driver = self.firewall
        self.plugin.init_is_complete = True

    def _default_rule(self):
        rule = DEFAULT_RULE
        rule['action'] = consts.FW_ACTION_ALLOW
        return rule

    def _fake_rules_v4(self, is_ingress=True):
        rule1 = {'enabled': True,
                 'action': 'allow',
                 'ip_version': 4,
                 'protocol': 'tcp',
                 'destination_port': '80',
                 'id': 'fake-fw-rule1',
                 'description': 'first rule'}
        rule2 = {'enabled': True,
                 'action': 'reject',
                 'ip_version': 4,
                 'protocol': 'tcp',
                 'destination_port': '22:24',
                 'source_port': '1:65535',
                 'id': 'fake-fw-rule2'}
        rule3 = {'enabled': True,
                 'action': 'deny',
                 'ip_version': 4,
                 'protocol': 'icmp',
                 'id': 'fake-fw-rule3'}
        rule4 = {'enabled': True,
                 'action': 'deny',
                 'ip_version': 4,
                 'id': 'fake-fw-rule4'}
        if is_ingress:
            # source ips are allowed
            rule1['source_ip_address'] = '10.24.4.2'
        else:
            # dest ips are allowed for egress rules
            rule1['destination_ip_address'] = '10.24.4.2'

        return [rule1, rule2, rule3, rule4]

    def _fake_translated_rules(self, nsx_port_id, is_ingress=True,
                               logged=False):
        # The expected translation of the rules in _fake_rules_v4
        service1 = {'l4_protocol': 'TCP',
                    'resource_type': 'L4PortSetNSService',
                    'destination_ports': ['80'],
                    'source_ports': []}
        rule1 = {'action': 'ALLOW',
                 'services': [{'service': service1}],
                 'sources': [{'target_id': '10.24.4.2',
                              'target_type': 'IPv4Address'}],
                 'display_name': 'Fwaas-fake-fw-rule1',
                 'notes': 'first rule'}
        if not is_ingress:
            rule1['destinations'] = rule1['sources']
            del rule1['sources']
        service2 = {'l4_protocol': 'TCP',
                    'resource_type': 'L4PortSetNSService',
                    'destination_ports': ['22-24'],
                    'source_ports': ['1-65535']}
        rule2 = {'action': 'DROP',  # Reject is replaced with deny
                 'services': [{'service': service2}],
                 'display_name': 'Fwaas-fake-fw-rule2'}
        service3_1 = {'resource_type': 'ICMPTypeNSService',
                      'protocol': 'ICMPv4'}
        service3_2 = {'resource_type': 'ICMPTypeNSService',
                      'protocol': 'ICMPv6'}
        rule3 = {'action': 'DROP',
                 # icmp is translated to icmp v4 & v6
                 'services': [{'service': service3_1},
                              {'service': service3_2}],
                 'display_name': 'Fwaas-fake-fw-rule3'}
        rule4 = {'action': 'DROP',
                 'display_name': 'Fwaas-fake-fw-rule4'}

        if nsx_port_id:
            if is_ingress:
                field = 'destinations'
                direction = 'IN'
            else:
                field = 'sources'
                direction = 'OUT'
            new_val = [{'target_id': nsx_port_id,
                        'target_type': 'LogicalSwitch'}]
            for rule in (rule1, rule2, rule3, rule4):
                rule[field] = new_val
                rule['direction'] = direction
        if logged:
            for rule in (rule1, rule2, rule3, rule4):
                rule['logged'] = logged
        return [rule1, rule2, rule3, rule4]

    def _fake_empty_firewall_group(self):
        fw_inst = {'id': FAKE_FW_ID,
                   'admin_state_up': True,
                   'tenant_id': 'tenant-uuid',
                   'ingress_rule_list': [],
                   'egress_rule_list': []}
        return fw_inst

    def _fake_firewall_group(self, rule_list, is_ingress=True,
                             admin_state_up=True):
        _rule_list = copy.deepcopy(rule_list)
        for rule in _rule_list:
            rule['position'] = str(_rule_list.index(rule))
        fw_inst = {'id': FAKE_FW_ID,
                   'admin_state_up': admin_state_up,
                   'tenant_id': 'tenant-uuid',
                   'ingress_rule_list': [],
                   'egress_rule_list': []}
        if is_ingress:
            fw_inst['ingress_rule_list'] = _rule_list
        else:
            fw_inst['egress_rule_list'] = _rule_list
        return fw_inst

    def _fake_firewall_group_with_admin_down(self, rule_list,
                                             is_ingress=True):
        return self._fake_firewall_group(
            rule_list, is_ingress=is_ingress, admin_state_up=False)

    def _fake_apply_list(self):
        router_inst = {'id': FAKE_ROUTER_ID}
        router_info_inst = mock.Mock()
        router_info_inst.router = router_inst
        router_info_inst.router_id = FAKE_ROUTER_ID
        apply_list = [(router_info_inst, FAKE_PORT_ID)]
        return apply_list

    def test_create_firewall_no_rules(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        port = {'id': FAKE_PORT_ID, 'network_id': FAKE_NET_ID}
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin, 'get_port',
                              return_value=port),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch("vmware_nsx.db.db.get_nsx_switch_and_port_id",
                       return_value=(FAKE_NSX_LS_ID, 0)),\
            mock.patch("vmware_nsxlib.v3.security.NsxLibFirewallSection."
                       "update") as update_fw:
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            # expecting 2 block rules for the logical port (egress & ingress)
            # and last default allow all rule
            expected_rules = [
                {'display_name': "Block port ingress",
                 'action': consts.FW_ACTION_DROP,
                 'destinations': [{'target_type': 'LogicalSwitch',
                                   'target_id': FAKE_NSX_LS_ID}],
                 'direction': 'IN'},
                {'display_name': "Block port egress",
                 'action': consts.FW_ACTION_DROP,
                 'sources': [{'target_type': 'LogicalSwitch',
                              'target_id': FAKE_NSX_LS_ID}],
                 'direction': 'OUT'},
                self._default_rule()
            ]
            update_fw.assert_called_once_with(
                MOCK_SECTION_ID,
                rules=expected_rules)

    def _setup_firewall_with_rules(self, func, is_ingress=True):
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4(is_ingress=is_ingress)
        firewall = self._fake_firewall_group(rule_list, is_ingress=is_ingress)
        port = {'id': FAKE_PORT_ID, 'network_id': FAKE_NET_ID}
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin, 'get_port',
                              return_value=port),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch("vmware_nsx.db.db.get_nsx_switch_and_port_id",
                       return_value=(FAKE_NSX_LS_ID, 0)),\
            mock.patch("vmware_nsxlib.v3.security.NsxLibFirewallSection."
                       "update") as update_fw:
            func('nsx', apply_list, firewall)
            expected_rules = self._fake_translated_rules(
                FAKE_NSX_LS_ID, is_ingress=is_ingress) + [
                {'display_name': "Block port ingress",
                 'action': consts.FW_ACTION_DROP,
                 'destinations': [{'target_type': 'LogicalSwitch',
                                   'target_id': FAKE_NSX_LS_ID}],
                 'direction': 'IN'},
                {'display_name': "Block port egress",
                 'action': consts.FW_ACTION_DROP,
                 'sources': [{'target_type': 'LogicalSwitch',
                              'target_id': FAKE_NSX_LS_ID}],
                 'direction': 'OUT'},
                self._default_rule()
            ]
            update_fw.assert_called_once_with(
                MOCK_SECTION_ID,
                rules=expected_rules)

    def test_create_firewall_with_ingress_rules(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group)

    def test_update_firewall_with_ingress_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group)

    def test_create_firewall_with_egress_rules(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group,
                                        is_ingress=False)

    def test_update_firewall_with_egress_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=False)

    def test_create_firewall_with_illegal_rules(self):
        """Use ingress rules as the egress list and verify failure"""
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4(is_ingress=True)
        firewall = self._fake_firewall_group(rule_list, is_ingress=False)
        self.assertRaises(exceptions.FirewallInternalDriverError,
                          self.firewall.create_firewall_group, 'nsx',
                          apply_list, firewall)

    def test_delete_firewall(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        port = {'id': FAKE_PORT_ID}
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=None),\
            mock.patch("vmware_nsx.db.db.get_nsx_switch_and_port_id",
                       return_value=(FAKE_NSX_LS_ID, 0)),\
            mock.patch("vmware_nsxlib.v3.security.NsxLibFirewallSection."
                       "update") as update_fw:
            self.firewall.delete_firewall_group('nsx', apply_list, firewall)
            update_fw.assert_called_once_with(
                MOCK_SECTION_ID,
                rules=[self._default_rule()])

    def test_create_firewall_with_admin_down(self):
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4()
        firewall = self._fake_firewall_group_with_admin_down(rule_list)
        with mock.patch("vmware_nsxlib.v3.security.NsxLibFirewallSection."
                        "update") as update_fw:
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            update_fw.assert_called_once_with(
                MOCK_SECTION_ID,
                rules=[self._default_rule()])

    def test_create_firewall_with_dhcp_relay(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        port = {'id': FAKE_PORT_ID, 'network_id': FAKE_NET_ID}
        relay_server = '1.1.1.1'
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin, 'get_port',
                              return_value=port),\
            mock.patch.object(self.plugin, '_get_port_relay_servers',
                              return_value=[relay_server]),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch("vmware_nsx.db.db.get_nsx_switch_and_port_id",
                       return_value=(FAKE_NSX_LS_ID, 0)),\
            mock.patch("vmware_nsxlib.v3.security.NsxLibFirewallSection."
                       "update") as update_fw:
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            # expecting 2 allow rules for the relay servers,
            # 2 block rules for the logical port (egress & ingress)
            # and last default allow all rule
            expected_rules = [
                {'display_name': "DHCP Relay ingress traffic",
                 'action': consts.FW_ACTION_ALLOW,
                 'destinations': [{'target_type': 'LogicalSwitch',
                                   'target_id': FAKE_NSX_LS_ID}],
                 'sources': [{'target_id': relay_server,
                              'target_type': 'IPv4Address'}],
                 'services': self.plugin._get_port_relay_services(),
                 'direction': 'IN'},
                {'display_name': "DHCP Relay egress traffic",
                 'action': consts.FW_ACTION_ALLOW,
                 'sources': [{'target_type': 'LogicalSwitch',
                              'target_id': FAKE_NSX_LS_ID}],
                 'destinations': [{'target_id': relay_server,
                                   'target_type': 'IPv4Address'}],
                 'services': self.plugin._get_port_relay_services(),
                 'direction': 'OUT'},
                {'display_name': "Block port ingress",
                 'action': consts.FW_ACTION_DROP,
                 'destinations': [{'target_type': 'LogicalSwitch',
                                   'target_id': FAKE_NSX_LS_ID}],
                 'direction': 'IN'},
                {'display_name': "Block port egress",
                 'action': consts.FW_ACTION_DROP,
                 'sources': [{'target_type': 'LogicalSwitch',
                              'target_id': FAKE_NSX_LS_ID}],
                 'direction': 'OUT'},
                self._default_rule()
            ]
            update_fw.assert_called_once_with(
                MOCK_SECTION_ID,
                rules=expected_rules)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron_fwaas.services.firewall.drivers import fwaas_base
from neutron_lib.api.definitions import constants as fwaas_consts
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.plugins import directory
from oslo_log import log as logging

from vmware_nsx.extensions import projectpluginmap
from vmware_nsxlib.v3 import nsx_constants as consts

LOG = logging.getLogger(__name__)
RULE_NAME_PREFIX = 'Fwaas-'
DEFAULT_RULE_NAME = 'Default LR Layer3 Rule'


class CommonEdgeFwaasV3Driver(fwaas_base.FwaasDriverBase):
    """Base class for NSX-V3 driver for Firewall As A Service - V1 & V2."""

    def __init__(self, driver_exception, driver_name):
        super(CommonEdgeFwaasV3Driver, self).__init__()
        self.driver_name = driver_name
        self.backend_support = True
        self.driver_exception = driver_exception
        registry.subscribe(
            self.check_backend_version,
            resources.PROCESS, events.BEFORE_SPAWN)
        self._core_plugin = None

    @property
    def core_plugin(self):
        """Get the NSX-V3 core plugin"""
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
            if self._core_plugin.is_tvd_plugin():
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_T)
                if not self._core_plugin:
                    # The nsx-t plugin was not initialized
                    return
            # make sure plugin init was completed
            if not self._core_plugin.init_is_complete:
                self._core_plugin.init_complete(None, None, {})
        return self._core_plugin

    @property
    def nsxlib(self):
        return self.core_plugin.nsxlib

    @property
    def nsx_firewall(self):
        return self.nsxlib.firewall_section

    @property
    def nsx_router(self):
        return self.nsxlib.logical_router

    def check_backend_version(self, resource, event, trigger, payload=None):
        if (self.core_plugin and
            not self.nsxlib.feature_supported(consts.FEATURE_ROUTER_FIREWALL)):
            # router firewall is not supported
            LOG.warning("FWaaS is not supported by the NSX backend (version "
                        "%s): Router firewall is not supported",
                        self.nsxlib.get_version())
            self.backend_support = False

    def should_apply_firewall_to_router(self, router_data):
        """Return True if the firewall rules should be added the router

        Right now the driver supports for all routers.
        """
        return True

    def _translate_action(self, fwaas_action, fwaas_rule_id):
        """Translate FWaaS action to NSX action"""
        if fwaas_action == fwaas_consts.FWAAS_ALLOW:
            return consts.FW_ACTION_ALLOW
        if fwaas_action == fwaas_consts.FWAAS_DENY:
            return consts.FW_ACTION_DROP
        if fwaas_action == fwaas_consts.FWAAS_REJECT:
            # reject is not supported by the nsx router firewall
            LOG.warning("Reject action is not supported by the NSX backend "
                        "for router firewall. Using %(action)s instead for "
                        "rule %(id)s",
                  {'action': consts.FW_ACTION_DROP,
                   'id': fwaas_rule_id})
            return consts.FW_ACTION_DROP
        # Unexpected action
        LOG.error("Unsupported FWAAS action %(action)s for rule %(id)s", {
            'action': fwaas_action, 'id': fwaas_rule_id})
        raise self.driver_exception(driver=self.driver_name)

    def _translate_cidr(self, cidr):
        return self.nsx_firewall.get_ip_cidr_reference(
            cidr,
            consts.IPV6 if netaddr.valid_ipv6(cidr) else consts.IPV4)

    def translate_addresses_to_target(self, cidrs):
        return [self._translate_cidr(ip) for ip in cidrs]

    @staticmethod
    def _translate_protocol(fwaas_protocol):
        """Translate FWaaS L4 protocol to NSX protocol"""
        if fwaas_protocol.lower() == 'tcp':
            return consts.TCP
        if fwaas_protocol.lower() == 'udp':
            return consts.UDP
        if fwaas_protocol.lower() == 'icmp':
            # This will cover icmpv6 too, when adding  the rule.
            return consts.ICMPV4

    @staticmethod
    def _translate_ports(ports):
        return [ports.replace(':', '-')]

    def _translate_services(self, fwaas_rule):
        l4_protocol = self._translate_protocol(fwaas_rule['protocol'])
        if l4_protocol in [consts.TCP, consts.UDP]:
            source_ports = []
            destination_ports = []
            if fwaas_rule.get('source_port'):
                source_ports = self._translate_ports(
                    fwaas_rule['source_port'])
            if fwaas_rule.get('destination_port'):
                destination_ports = self._translate_ports(
                    fwaas_rule['destination_port'])

            return [self.nsx_firewall.get_nsservice(
                consts.L4_PORT_SET_NSSERVICE,
                l4_protocol=l4_protocol,
                source_ports=source_ports,
                destination_ports=destination_ports)]
        elif l4_protocol == consts.ICMPV4:
            # Add both icmp v4 & v6 services
            return [
                self.nsx_firewall.get_nsservice(
                    consts.ICMP_TYPE_NSSERVICE,
                    protocol=consts.ICMPV4),
                self.nsx_firewall.get_nsservice(
                    consts.ICMP_TYPE_NSSERVICE,
                    protocol=consts.ICMPV6),
            ]

    def _translate_rules(self, fwaas_rules, replace_src=None,
                         replace_dest=None, logged=False):
        translated_rules = []
        for rule in fwaas_rules:
            nsx_rule = {}
            if not rule['enabled']:
                # skip disabled rules
                continue
            # Make sure the rule has a name, and it starts with the prefix
            # (backend max name length is 255)
            if rule.get('name'):
                name = RULE_NAME_PREFIX + rule['name']
            else:
                name = RULE_NAME_PREFIX + rule['id']
            nsx_rule['display_name'] = name[:255]
            if rule.get('description'):
                nsx_rule['notes'] = rule['description']
            nsx_rule['action'] = self._translate_action(
                rule['action'], rule['id'])
            if replace_dest:
                # set this value as the destination logical switch,
                # and set the rule to ingress
                nsx_rule['destinations'] = [{'target_type': 'LogicalSwitch',
                                             'target_id': replace_dest}]
                nsx_rule['direction'] = 'IN'
            elif rule.get('destination_ip_address'):
                nsx_rule['destinations'] = self.translate_addresses_to_target(
                    [rule['destination_ip_address']])
            if replace_src:
                # set this value as the source logical switch,
                # and set the rule to egress
                nsx_rule['sources'] = [{'target_type': 'LogicalSwitch',
                                        'target_id': replace_src}]
                nsx_rule['direction'] = 'OUT'
            elif rule.get('source_ip_address'):
                nsx_rule['sources'] = self.translate_addresses_to_target(
                    [rule['source_ip_address']])
            if rule.get('protocol'):
                nsx_rule['services'] = self._translate_services(rule)
            if logged:
                nsx_rule['logged'] = logged
            translated_rules.append(nsx_rule)

        return translated_rules

    def validate_backend_version(self):
        # prevent firewall actions if the backend does not support it
        if not self.backend_support:
            LOG.error("The NSX backend does not support router firewall")
            raise self.driver_exception(driver=self.driver_name)

    def get_default_backend_rule(self, section_id, allow_all=True):
        # Add default allow all rule
        old_default_rule = self.nsx_firewall.get_default_rule(
            section_id)
        return {
            'display_name': DEFAULT_RULE_NAME,
            'action': (consts.FW_ACTION_ALLOW if allow_all
                       else consts.FW_ACTION_DROP),
            'is_default': True,
            'id': old_default_rule['id'] if old_default_rule else 0}



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3 import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc

LOG = logging.getLogger(__name__)


class EdgeL7PolicyManager(base_mgr.Nsxv3LoadbalancerBaseManager):
    @log_helpers.log_method_call
    def __init__(self):
        super(EdgeL7PolicyManager, self).__init__()

    @log_helpers.log_method_call
    def _update_policy_position(self, vs_id, rule_id, position):
        vs_client = self.core_plugin.nsxlib.load_balancer.virtual_server
        vs = vs_client.get(vs_id)
        lb_rules = vs.get('rule_ids', [])
        if rule_id in lb_rules:
            lb_rules.remove(rule_id)
        if len(lb_rules) < position:
            lb_rules.append(rule_id)
        else:
            lb_rules.insert(position - 1, rule_id)
        vs_client.update(vs_id, rule_ids=lb_rules)

    @log_helpers.log_method_call
    def create(self, context, policy):
        lb_id = policy.listener.loadbalancer_id
        listener_id = policy.listener_id
        rule_client = self.core_plugin.nsxlib.load_balancer.rule
        tags = lb_utils.get_tags(self.core_plugin, policy.id,
                                 lb_const.LB_L7POLICY_TYPE,
                                 policy.tenant_id, context.project_name)

        binding = nsx_db.get_nsx_lbaas_listener_binding(
            context.session, lb_id, listener_id)
        if not binding:
            self.lbv2_driver.l7policy.failed_completion(context, policy)
            msg = _('Cannot find nsx lbaas binding for listener '
                    '%(listener_id)s') % {'listener_id': listener_id}
            raise n_exc.BadRequest(resource='lbaas-l7policy-create', msg=msg)

        vs_id = binding['lb_vs_id']
        rule_body = lb_utils.convert_l7policy_to_lb_rule(context, policy)
        try:
            lb_rule = rule_client.create(tags=tags, **rule_body)
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                self.lbv2_driver.l7policy.failed_completion(context, policy)
                LOG.error('Failed to create lb rule at NSX backend')
        try:
            self._update_policy_position(vs_id, lb_rule['id'],
                                         policy.position)
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                self.lbv2_driver.l7policy.failed_completion(context, policy)
                LOG.error('Failed to add rule %(rule)% to virtual server '
                          '%(vs)s at NSX backend', {'rule': lb_rule['id'],
                                                    'vs': vs_id})

        nsx_db.add_nsx_lbaas_l7policy_binding(
            context.session, policy.id, lb_rule['id'], vs_id)
        self.lbv2_driver.l7policy.successful_completion(context, policy)

    @log_helpers.log_method_call
    def update(self, context, old_policy, new_policy):
        rule_client = self.core_plugin.nsxlib.load_balancer.rule
        binding = nsx_db.get_nsx_lbaas_l7policy_binding(context.session,
                                                        old_policy.id)
        if not binding:
            self.lbv2_driver.l7rule.failed_completion(context, new_policy)
            msg = _('Cannot find nsx lbaas binding for policy '
                    '%(policy_id)s') % {'policy_id': old_policy.id}
            raise n_exc.BadRequest(resource='lbaas-l7policy-update', msg=msg)

        vs_id = binding['lb_vs_id']
        lb_rule_id = binding['lb_rule_id']
        rule_body = lb_utils.convert_l7policy_to_lb_rule(context, new_policy)
        try:
            rule_client.update(lb_rule_id, **rule_body)
            if new_policy.position != old_policy.position:
                self._update_policy_position(vs_id, lb_rule_id,
                                             new_policy.position)

        except Exception as e:
            with excutils.save_and_reraise_exception():
                self.lbv2_driver.l7policy.failed_completion(context,
                                                            new_policy)
                LOG.error('Failed to update L7policy %(policy)s: '
                          '%(err)s', {'policy': old_policy.id, 'err': e})

        self.lbv2_driver.l7policy.successful_completion(context, new_policy)

    @log_helpers.log_method_call
    def delete(self, context, policy):
        vs_client = self.core_plugin.nsxlib.load_balancer.virtual_server
        rule_client = self.core_plugin.nsxlib.load_balancer.rule
        binding = nsx_db.get_nsx_lbaas_l7policy_binding(context.session,
                                                        policy.id)
        if binding:
            vs_id = binding['lb_vs_id']
            rule_id = binding['lb_rule_id']
            try:
                # Update virtual server to remove lb rule
                vs_client.remove_rule(vs_id, rule_id)
                rule_client.delete(rule_id)
            except nsxlib_exc.ResourceNotFound:
                LOG.warning('LB rule %(rule)s is not found on NSX',
                            {'rule': rule_id})
            except nsxlib_exc.ManagerError:
                self.lbv2_driver.l7policy.failed_completion(
                    context, policy)
                msg = (_('Failed to delete lb rule: %(rule)s') %
                       {'rule': rule_id})
                raise n_exc.BadRequest(resource='lbaas-l7policy-delete',
                                       msg=msg)
            nsx_db.delete_nsx_lbaas_l7policy_binding(
                context.session, policy.id)

        self.lbv2_driver.l7policy.successful_completion(
            context, policy, delete=True)



# Copyright (c) 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from vmware_nsx.tests.unit.nsx_v3 import test_plugin

from neutron_lib.api import attributes
from neutron_lib.plugins import directory
from oslo_config import cfg


class TestApiReplay(test_plugin.NsxV3PluginTestCaseMixin):

    def setUp(self, plugin=None, ext_mgr=None, service_plugins=None):
        # enables api_replay_mode for these tests
        cfg.CONF.set_override('api_replay_mode', True)

        super(TestApiReplay, self).setUp()

    def tearDown(self):
        # disables api_replay_mode for these tests
        cfg.CONF.set_override('api_replay_mode', False)

        # remove the extension from the plugin
        directory.get_plugin().supported_extension_aliases.remove(
            'api-replay')

        # Revert the attributes map back to normal
        for attr_name in ('ports', 'networks', 'security_groups',
                          'security_group_rules', 'routers', 'policies'):
            attr_info = attributes.RESOURCES[attr_name]
            attr_info['id']['allow_post'] = False

        super(TestApiReplay, self).tearDown()

    def test_create_port_specify_id(self):
        specified_network_id = '555e762b-d7a1-4b44-b09b-2a34ada56c9f'
        specified_port_id = 'e55e762b-d7a1-4b44-b09b-2a34ada56c9f'
        network_res = self._create_network(self.fmt,
                                           'test-network',
                                           True,
                                           arg_list=('id',),
                                           id=specified_network_id)
        network = self.deserialize(self.fmt, network_res)
        self.assertEqual(specified_network_id, network['network']['id'])
        port_res = self._create_port(self.fmt,
                                     network['network']['id'],
                                     arg_list=('id',),
                                     id=specified_port_id)
        port = self.deserialize(self.fmt, port_res)
        self.assertEqual(specified_port_id, port['port']['id'])



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import hashlib

from cryptography import fernet
from oslo_config import cfg
from oslo_log import log as logging

from vmware_nsx.db import db as nsx_db

LOG = logging.getLogger(__name__)
NSX_OPENSTACK_IDENTITY = "com.vmware.nsx.openstack"

# 32-byte base64-encoded secret for symmetric password encryption
# generated on init based on password provided in configuration
_SECRET = None


def reset_secret():
    global _SECRET
    _SECRET = None


def generate_secret_from_password(password):
    m = hashlib.md5()
    m.update(password.encode('ascii'))
    return base64.b64encode(m.hexdigest().encode('ascii'))


def symmetric_encrypt(secret, plaintext):
    if not isinstance(plaintext, bytes):
        plaintext = plaintext.encode('ascii')
    return fernet.Fernet(secret).encrypt(plaintext).decode('ascii')


def symmetric_decrypt(secret, ciphertext):
    if not isinstance(ciphertext, bytes):
        ciphertext = ciphertext.encode('ascii')
    return fernet.Fernet(secret).decrypt(ciphertext).decode('ascii')


class DbCertificateStorageDriver(object):
    """Storage for certificate and private key in neutron DB"""
    def __init__(self, context):
        global _SECRET
        self._context = context
        if cfg.CONF.nsx_v3.nsx_client_cert_pk_password and not _SECRET:
            _SECRET = generate_secret_from_password(
                    cfg.CONF.nsx_v3.nsx_client_cert_pk_password)

    def store_cert(self, purpose, certificate, private_key):
        # encrypt private key
        if _SECRET:
            private_key = symmetric_encrypt(_SECRET, private_key)

        nsx_db.save_certificate(self._context.session, purpose,
                                certificate, private_key)

    def get_cert(self, purpose):
        cert, private_key = nsx_db.get_certificate(self._context.session,
                                                   purpose)
        if _SECRET and private_key:
            try:
                # Encrypted PK is stored in DB as string, while fernet expects
                # bytearray.
                private_key = symmetric_decrypt(_SECRET, private_key)
            except fernet.InvalidToken:
                # unable to decrypt - probably due to change of password
                # cert and PK are useless, need to delete them
                LOG.error("Unable to decrypt private key, possibly due "
                          "to change of password. Certificate needs to be "
                          "regenerated")
                self.delete_cert(purpose)
                return None, None

        return cert, private_key

    def delete_cert(self, purpose):
        return nsx_db.delete_certificate(self._context.session, purpose)


class DummyCertificateStorageDriver(object):
    """Dummy driver API implementation

    Used for external certificate import scenario
    (nsx_client_cert_storage == None)
    """

    def store_cert(self, purpose, certificate, private_key):
        pass

    def get_cert(self, purpose):
        return None, None

    def delete_cert(self, purpose):
        pass



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_lib import exceptions as n_exc

from vmware_nsx.common import locking
from vmware_nsx.db import nsxv_db
from vmware_nsx.plugins.nsx_v.vshield.common import exceptions as nsxv_exc
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v import lbaas_common as lb_common

LOG = logging.getLogger(__name__)


class EdgeHealthMonitorManager(base_mgr.EdgeLoadbalancerBaseManager):

    @log_helpers.log_method_call
    def _convert_lbaas_monitor(self, hm):
        """
        Transform OpenStack health monitor dict to NSXv health monitor dict.
        """
        mon = {
            'type': lb_const.HEALTH_MONITOR_MAP.get(hm.type, 'icmp'),
            'interval': hm.delay,
            'timeout': hm.timeout,
            'maxRetries': hm.max_retries,
            'name': hm.id}

        if hm.http_method:
            mon['method'] = hm.http_method

        if hm.url_path:
            mon['url'] = hm.url_path
        return mon

    @log_helpers.log_method_call
    def __init__(self, vcns_driver):
        super(EdgeHealthMonitorManager, self).__init__(vcns_driver)

    @log_helpers.log_method_call
    def create(self, context, hm):
        lb_id = hm.pool.loadbalancer_id
        lb_binding = nsxv_db.get_nsxv_lbaas_loadbalancer_binding(
            context.session, lb_id)
        edge_id = lb_binding['edge_id']

        pool_binding = nsxv_db.get_nsxv_lbaas_pool_binding(
            context.session, lb_id, hm.pool.id)
        if not pool_binding:
            self.lbv2_driver.health_monitor.failed_completion(
                context, hm)
            msg = _('Failed to create health monitor on edge: %s. '
                    'Binding not found') % edge_id
            LOG.error(msg)
            raise n_exc.BadRequest(resource='edge-lbaas', msg=msg)

        edge_pool_id = pool_binding['edge_pool_id']

        hm_binding = nsxv_db.get_nsxv_lbaas_monitor_binding(
            context.session, lb_id, hm.pool.id, hm.id, edge_id)
        edge_mon_id = None

        if hm_binding:
            edge_mon_id = hm_binding['edge_mon_id']
        else:
            edge_monitor = self._convert_lbaas_monitor(hm)
            try:
                with locking.LockManager.get_lock(edge_id):
                    h = self.vcns.create_health_monitor(edge_id,
                                                        edge_monitor)[0]
                    edge_mon_id = lb_common.extract_resource_id(h['location'])

                nsxv_db.add_nsxv_lbaas_monitor_binding(
                    context.session, lb_id, hm.pool.id, hm.id, edge_id,
                    edge_mon_id)

            except nsxv_exc.VcnsApiException:
                with excutils.save_and_reraise_exception():
                    self.lbv2_driver.health_monitor.failed_completion(
                        context, hm)
                    LOG.error('Failed to create health monitor on edge: %s',
                              edge_id)

        try:
            # Associate monitor with Edge pool
            with locking.LockManager.get_lock(edge_id):
                edge_pool = self.vcns.get_pool(edge_id, edge_pool_id)[1]
                if edge_pool.get('monitorId'):
                    edge_pool['monitorId'].append(edge_mon_id)
                else:
                    edge_pool['monitorId'] = [edge_mon_id]

                self.vcns.update_pool(edge_id, edge_pool_id, edge_pool)

        except nsxv_exc.VcnsApiException:
            with excutils.save_and_reraise_exception():
                self.lbv2_driver.health_monitor.failed_completion(context, hm)
                LOG.error(
                    'Failed to create health monitor on edge: %s',
                    edge_id)

        self.lbv2_driver.health_monitor.successful_completion(context, hm)

    @log_helpers.log_method_call
    def update(self, context, old_hm, new_hm):
        lb_id = new_hm.pool.loadbalancer_id
        lb_binding = nsxv_db.get_nsxv_lbaas_loadbalancer_binding(
            context.session, lb_id)

        edge_id = lb_binding['edge_id']

        hm_binding = nsxv_db.get_nsxv_lbaas_monitor_binding(
            context.session, lb_id, new_hm.pool.id, new_hm.id, edge_id)

        edge_monitor = self._convert_lbaas_monitor(new_hm)

        try:
            with locking.LockManager.get_lock(edge_id):
                self.vcns.update_health_monitor(edge_id,
                                                hm_binding['edge_mon_id'],
                                                edge_monitor)

        except nsxv_exc.VcnsApiException:
            with excutils.save_and_reraise_exception():
                self.lbv2_driver.health_monitor.failed_completion(context,
                                                                  new_hm)
                LOG.error('Failed to update monitor on edge: %s', edge_id)

        self.lbv2_driver.health_monitor.successful_completion(context, new_hm)

    @log_helpers.log_method_call
    def delete(self, context, hm):
        pool_id = hm.pool.id
        lb_id = hm.pool.loadbalancer_id
        lb_binding = nsxv_db.get_nsxv_lbaas_loadbalancer_binding(
            context.session, lb_id)
        edge_id = lb_binding['edge_id']

        pool_binding = nsxv_db.get_nsxv_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if not pool_binding:
            nsxv_db.del_nsxv_lbaas_monitor_binding(
                context.session, lb_id, pool_id, hm.id, edge_id)
            self.lbv2_driver.health_monitor.successful_completion(
                context, hm, delete=True)
            return

        edge_pool_id = pool_binding['edge_pool_id']

        hm_binding = nsxv_db.get_nsxv_lbaas_monitor_binding(
            context.session, lb_id, pool_id, hm.id, edge_id)

        edge_pool = self.vcns.get_pool(edge_id, edge_pool_id)[1]
        if hm_binding['edge_mon_id'] in edge_pool['monitorId']:
            edge_pool['monitorId'].remove(hm_binding['edge_mon_id'])

            try:
                with locking.LockManager.get_lock(edge_id):
                    self.vcns.update_pool(edge_id, edge_pool_id, edge_pool)
            except nsxv_exc.VcnsApiException:
                with excutils.save_and_reraise_exception():
                    self.lbv2_driver.health_monitor.failed_completion(context,
                                                                      hm)
                    LOG.error('Failed to delete monitor mapping on edge: %s',
                              edge_id)

        # If this monitor is not used on this edge anymore, delete it
        if not edge_pool['monitorId']:
            try:
                with locking.LockManager.get_lock(edge_id):
                    self.vcns.delete_health_monitor(hm_binding['edge_id'],
                                                    hm_binding['edge_mon_id'])
            except nsxv_exc.VcnsApiException:
                with excutils.save_and_reraise_exception():
                    self.lbv2_driver.health_monitor.failed_completion(context,
                                                                      hm)
                    LOG.error('Failed to delete monitor on edge: %s', edge_id)

        nsxv_db.del_nsxv_lbaas_monitor_binding(
            context.session, lb_id, pool_id, hm.id, edge_id)
        self.lbv2_driver.health_monitor.successful_completion(
            context, hm, delete=True)



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron.ipam import exceptions as ipam_exc
from neutron.ipam import subnet_alloc

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx import utils as tvd_utils
from vmware_nsx.services.ipam.common import driver as common_driver
from vmware_nsx.services.ipam.nsx_v import driver as v_driver
from vmware_nsx.services.ipam.nsx_v3 import driver as t_driver

LOG = logging.getLogger(__name__)


class NsxTvdIpamDriver(subnet_alloc.SubnetAllocator,
                       common_driver.NsxIpamBase):
    """IPAM Driver For NSX-TVD plugin."""

    def __init__(self, subnetpool, context):
        super(NsxTvdIpamDriver, self).__init__(subnetpool, context)
        # initialize the different drivers
        self.drivers = {}
        try:
            self.drivers[projectpluginmap.NsxPlugins.NSX_T] = (
                t_driver.Nsxv3IpamDriver(subnetpool, context))
        except Exception as e:
            LOG.warning("NsxTvdIpamDriver failed to initialize the NSX-T "
                        "driver %s", e)
            self.drivers[projectpluginmap.NsxPlugins.NSX_T] = None
        try:
            self.drivers[projectpluginmap.NsxPlugins.NSX_V] = (
                v_driver.NsxvIpamDriver(subnetpool, context))
        except Exception as e:
            LOG.warning("NsxTvdIpamDriver failed to initialize the NSX-V "
                        "driver %s", e)
            self.drivers[projectpluginmap.NsxPlugins.NSX_V] = None

    def get_T_driver(self):
        return self.drivers[projectpluginmap.NsxPlugins.NSX_T]

    def get_V_driver(self):
        return self.drivers[projectpluginmap.NsxPlugins.NSX_V]

    def _get_driver_for_project(self, project):
        plugin_type = tvd_utils.get_tvd_plugin_type_for_project(project)
        if not self.drivers.get(plugin_type):
            LOG.error("Project %(project)s with plugin %(plugin)s has no "
                      "support for IPAM", {'project': project,
                                           'plugin': plugin_type})
            raise ipam_exc.IpamValueInvalid(
                msg="IPAM driver not found")
        return self.drivers[plugin_type]

    def allocate_subnet(self, subnet_request):
        d = self._get_driver_for_project(subnet_request.tenant_id)
        return d.allocate_subnet(subnet_request)

    def update_subnet(self, subnet_request):
        d = self._get_driver_for_project(subnet_request.tenant_id)
        return d.update_subnet(subnet_request)

    def remove_subnet(self, subnet_id):
        d = self._get_driver_for_project(self._context.tenant_id)
        return d.remove_subnet(subnet_id)

    def get_subnet(self, subnet_id):
        d = self._get_driver_for_project(self._context.tenant_id)
        return d.get_subnet(subnet_id)

    def get_subnet_request_factory(self):
        d = self._get_driver_for_project(self._context.tenant_id)
        return d.get_subnet_request_factory()



# Copyright 2014 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_vmware import api
from oslo_vmware import exceptions as oslo_vmware_exc

from vmware_nsx._i18n import _

dvs_opts = [
    cfg.StrOpt('host_ip',
               help='Hostname or IP address for connection to VMware vCenter '
                    'host.'),
    cfg.PortOpt('host_port', default=443,
                help='Port for connection to VMware vCenter host.'),
    cfg.StrOpt('host_username',
               help='Username for connection to VMware vCenter host.'),
    cfg.StrOpt('host_password',
               help='Password for connection to VMware vCenter host.',
               secret=True),
    cfg.FloatOpt('task_poll_interval',
                 default=0.5,
                 help='The interval used for polling of remote tasks.'),
    cfg.StrOpt('ca_file',
               help='Specify a CA bundle file to use in verifying the '
                    'vCenter server certificate.'),
    cfg.BoolOpt('insecure',
                default=False,
                help='If true, the vCenter server certificate is not '
                     'verified. If false, then the default CA truststore is '
                     'used for verification. This option is ignored if '
                     '"ca_file" is set.'),
    cfg.IntOpt('api_retry_count',
               default=10,
               help='The number of times we retry on failures, e.g., '
                    'socket error, etc.'),
    cfg.StrOpt('dvs_name',
               help='The name of the preconfigured DVS.'),
    cfg.StrOpt('metadata_mode',
               help=_("This value should not be set. It is just required for "
                      "ensuring that the DVS plugin works with the generic "
                      "NSX metadata code")),
]

CONF = cfg.CONF
CONF.register_opts(dvs_opts, 'dvs')


# Create and register exceptions not in oslo.vmware
class DvsOperationBulkFault(oslo_vmware_exc.VimException):
    msg_fmt = _("Cannot complete a DVS operation for one or more members.")


def dvs_register_exceptions():
    oslo_vmware_exc.register_fault_class('DvsOperationBulkFault',
                                         DvsOperationBulkFault)


def dvs_is_enabled(dvs_id=None):
    """Returns the configured DVS status."""
    return bool(CONF.dvs.host_ip and CONF.dvs.host_username and
                CONF.dvs.host_password and (dvs_id or CONF.dvs.dvs_name))


def dvs_create_session():
    return api.VMwareAPISession(CONF.dvs.host_ip,
                                CONF.dvs.host_username,
                                CONF.dvs.host_password,
                                CONF.dvs.api_retry_count,
                                CONF.dvs.task_poll_interval,
                                port=CONF.dvs.host_port,
                                cacert=CONF.dvs.ca_file,
                                insecure=CONF.dvs.insecure)


def dvs_name_get():
    return CONF.dvs.dvs_name



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron_lib import constants as nl_constants

from vmware_nsx.db import db as nsx_db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.fwaas.common import fwaas_callbacks_v2 as \
    com_callbacks
from vmware_nsx.services.fwaas.nsx_tv import edge_fwaas_driver_v2 as tv_driver

LOG = logging.getLogger(__name__)


class Nsxv3FwaasCallbacksV2(com_callbacks.NsxFwaasCallbacksV2):
    """NSX-V3 RPC callbacks for Firewall As A Service - V2."""

    def __init__(self):
        super(Nsxv3FwaasCallbacksV2, self).__init__()
        # update the fwaas driver in case of TV plugin
        self.internal_driver = None
        if self.fwaas_enabled:
            if self.fwaas_driver.driver_name == tv_driver.FWAAS_DRIVER_NAME:
                self.internal_driver = self.fwaas_driver.get_T_driver()
            else:
                self.internal_driver = self.fwaas_driver

    @property
    def plugin_type(self):
        return projectpluginmap.NsxPlugins.NSX_T

    def should_apply_firewall_to_router(self, context, router_id):
        """Return True if the FWaaS rules should be added to this router."""
        if not super(Nsxv3FwaasCallbacksV2,
                     self).should_apply_firewall_to_router(context,
                                                           router_id):
            return False

        # get all the relevant router info
        ctx_elevated = context.elevated()
        router_data = self.core_plugin.get_router(ctx_elevated, router_id)
        if not router_data:
            LOG.error("Couldn't read router %s data", router_id)
            return False

        # Check if the FWaaS driver supports this router
        if not self.internal_driver.should_apply_firewall_to_router(
            router_data):
            return False

        return True

    def get_port_rules(self, nsx_ls_id, fwg, plugin_rules):
        return self.internal_driver.get_port_translated_rules(
            nsx_ls_id, fwg, plugin_rules)

    def update_router_firewall(self, context, nsxlib, router_id,
                               router_interfaces, nsx_router_id, section_id):
        """Rewrite all the FWaaS v2 rules in the router edge firewall

        This method should be called on FWaaS updates, and on router
        interfaces changes.
        """
        fw_rules = []
        # Add firewall rules per port attached to a firewall group
        for port in router_interfaces:
            nsx_ls_id, _nsx_port_id = nsx_db.get_nsx_switch_and_port_id(
                context.session, port['id'])

            # Check if this port has a firewall
            fwg = self.get_port_fwg(context, port['id'])
            if fwg:
                # Add plugin additional allow rules
                plugin_rules = self.core_plugin.get_extra_fw_rules(
                    context, router_id, port['id'])

                # add the FWaaS rules for this port
                # ingress/egress firewall rules + default ingress/egress drop
                # rule for this port
                fw_rules.extend(self.get_port_rules(nsx_ls_id, fwg,
                                                    plugin_rules))

        # add a default allow-all rule to all other traffic & ports
        fw_rules.append(self.internal_driver.get_default_backend_rule(
            section_id, allow_all=True))

        # update the backend router firewall
        nsxlib.firewall_section.update(section_id, rules=fw_rules)

    def delete_port(self, context, port_id):
        # Mark the FW group as inactive if this is the last port
        fwg = self.get_port_fwg(context, port_id)
        if (fwg and fwg.get('status') == nl_constants.ACTIVE and
            len(fwg.get('ports', [])) <= 1):
            self.fwplugin_rpc.set_firewall_group_status(
                context, fwg['id'], nl_constants.INACTIVE)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.common import ovs_lib
from neutron.agent.linux import dhcp

LOG = logging.getLogger(__name__)

OPTS = [
    cfg.StrOpt('dvs_integration_bridge',
               default='br-dvs',
               help=_('Name of Open vSwitch bridge to use for DVS networks')),
    cfg.StrOpt('dhcp_override_mac',
               help=_('Override the MAC address of the DHCP interface')),
]

cfg.CONF.register_opts(OPTS)


class DeviceManager(dhcp.DeviceManager):

    def plug(self, network, port, interface_name):
        mac_address = (cfg.CONF.dhcp_override_mac
                       if cfg.CONF.dhcp_override_mac
                       else port.mac_address)
        self.driver.plug(network.id,
                         port.id,
                         interface_name,
                         mac_address,
                         namespace=network.namespace,
                         mtu=network.get('mtu'),
                         bridge=cfg.CONF.dvs_integration_bridge)
        vlan_tag = getattr(network, 'provider:segmentation_id',
                           None)
        # Treat vlans
        if vlan_tag and vlan_tag != 0:
            br_dvs = ovs_lib.OVSBridge(self.conf.dvs_integration_bridge)
            # When ovs_use_veth is set to True, the DEV_NAME_PREFIX
            # will be changed from 'tap' to 'ns-' in
            # OVSInterfaceDriver
            dvs_port_name = interface_name.replace('ns-', 'tap')
            br_dvs.set_db_attribute("Port", dvs_port_name, "tag", vlan_tag)

    def unplug(self, device_name, network):
        self.driver.unplug(
            device_name, bridge=cfg.CONF.dvs_integration_bridge,
            namespace=network.namespace)


class Dnsmasq(dhcp.Dnsmasq):

    def __init__(self, conf, network, process_monitor,
                 version=None, plugin=None):
        super(Dnsmasq, self).__init__(conf, network, process_monitor,
                                      version=version, plugin=plugin)
        # Using the DeviceManager that enables us to directly plug the OVS
        LOG.debug("Using the DVS DeviceManager")
        self.device_manager = DeviceManager(conf, plugin)



# Copyright 2014 VMware, Inc.  All rights reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from vmware_nsx.common import nsxv_constants
from vmware_nsx.db import (
    distributedrouter as dist_rtr)
from vmware_nsx.extensions import routertype as rt_rtr


class RouterType_mixin(dist_rtr.DistributedRouter_mixin):
    """Mixin class to enable Router type support."""

    nsx_attributes = (
        dist_rtr.DistributedRouter_mixin.nsx_attributes + [{
            'name': rt_rtr.ROUTER_TYPE,
            'default': nsxv_constants.SHARED
        }])



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_lib.exceptions import firewall_v2 as exceptions

from vmware_nsx.services.fwaas.nsx_v3 import edge_fwaas_driver_base \
    as base_driver
from vmware_nsxlib.v3 import nsx_constants as consts

LOG = logging.getLogger(__name__)
FWAAS_DRIVER_NAME = 'Fwaas V2 NSX-V3 driver'


class EdgeFwaasV3DriverV2(base_driver.CommonEdgeFwaasV3Driver):
    """NSX-V3 driver for Firewall As A Service - V2."""

    def __init__(self):
        exception_cls = exceptions.FirewallInternalDriverError
        super(EdgeFwaasV3DriverV2, self).__init__(exception_cls,
                                                  FWAAS_DRIVER_NAME)

    @log_helpers.log_method_call
    def create_firewall_group(self, agent_mode, apply_list, firewall_group):
        """Create the Firewall with a given policy. """
        self._validate_firewall_group(firewall_group)
        self._update_backend_routers(apply_list, firewall_group['id'])

    @log_helpers.log_method_call
    def update_firewall_group(self, agent_mode, apply_list, firewall_group):
        """Remove previous policy and apply the new policy."""
        self._validate_firewall_group(firewall_group)
        self._update_backend_routers(apply_list, firewall_group['id'])

    @log_helpers.log_method_call
    def delete_firewall_group(self, agent_mode, apply_list, firewall_group):
        """Delete firewall.

        Removes rules created by this instance from the backend firewall
        And add the default allow rule.
        """
        self._update_backend_routers(apply_list, firewall_group['id'])

    @log_helpers.log_method_call
    def apply_default_policy(self, agent_mode, apply_list, firewall_group):
        """Apply the default policy (deny all).

        The backend firewall always has this policy (=deny all) as default,
        so we only need to delete the current rules.
        """
        self._update_backend_routers(apply_list, firewall_group['id'])

    def _update_backend_routers(self, apply_list, fwg_id):
        """Update all the affected router on the backend"""
        self.validate_backend_version()
        LOG.info("Updating routers firewall for firewall group %s", fwg_id)
        context = n_context.get_admin_context()
        routers = set()
        # the apply_list is a list of tuples: routerInfo, port-id
        for router_info, port_id in apply_list:
            # Skip unsupported routers
            if not self.should_apply_firewall_to_router(router_info.router):
                continue
            routers.add(router_info.router_id)

        # update each router once
        for router_id in routers:
            self.core_plugin.update_router_firewall(context, router_id)

    def get_port_translated_rules(self, nsx_ls_id, firewall_group,
                                  plugin_rules):
        """Return the list of translated rules per port"""
        port_rules = []
        # TODO(asarfaty): get this value from the firewall group extensions
        logged = False
        # Add the firewall group ingress/egress rules only if the fw is up
        if firewall_group['admin_state_up']:
            port_rules.extend(self._translate_rules(
                firewall_group['ingress_rule_list'],
                replace_dest=nsx_ls_id,
                logged=logged))
            port_rules.extend(self._translate_rules(
                firewall_group['egress_rule_list'],
                replace_src=nsx_ls_id,
                logged=logged))

        # Add the per-port plugin rules
        if plugin_rules and isinstance(plugin_rules, list):
            port_rules.extend(plugin_rules)

        # Add ingress/egress block rules for this port
        port_rules.extend([
            {'display_name': "Block port ingress",
             'action': consts.FW_ACTION_DROP,
             'destinations': [{'target_type': 'LogicalSwitch',
                               'target_id': nsx_ls_id}],
             'direction': 'IN'},
            {'display_name': "Block port egress",
             'action': consts.FW_ACTION_DROP,
             'sources': [{'target_type': 'LogicalSwitch',
                          'target_id': nsx_ls_id}],
             'direction': 'OUT'}])

        return port_rules

    def _validate_firewall_group(self, firewall_group):
        """Validate the rules in the firewall group"""
        for rule in firewall_group['egress_rule_list']:
            if rule.get('source_ip_address'):
                # this rule cannot be used as egress rule
                LOG.error("Rule %(id)s cannot be used in an egress "
                          "policy because it has a source",
                          {'id': rule['id']})
                raise exceptions.FirewallInternalDriverError(
                    driver=FWAAS_DRIVER_NAME)
        for rule in firewall_group['ingress_rule_list']:
            if rule.get('destination_ip_address'):
                # this rule cannot be used as ingress rule
                LOG.error("Rule %(id)s cannot be used in an ingress "
                          "policy because it has a destination",
                          {'id': rule['id']})
                raise exceptions.FirewallInternalDriverError(
                    driver=FWAAS_DRIVER_NAME)



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.extensions import securitygroup

from neutron_lib.api import extensions
from neutron_lib import constants

LOCAL_IP_PREFIX = 'local_ip_prefix'

RESOURCE_ATTRIBUTE_MAP = {
    'security_group_rules': {
        LOCAL_IP_PREFIX: {
            'allow_post': True,
            'allow_put': False,
            'convert_to': securitygroup.convert_ip_prefix_to_cidr,
            'default': constants.ATTR_NOT_SPECIFIED,
            'enforce_policy': True,
            'is_visible': True}
    }
}


class Secgroup_rule_local_ip_prefix(extensions.ExtensionDescriptor):
    """Extension class to add support for specifying local-ip-prefix in a
    security-group rule.
    """

    @classmethod
    def get_name(cls):
        return "Security Group rule local ip prefix"

    @classmethod
    def get_alias(cls):
        return "secgroup-rule-local-ip-prefix"

    @classmethod
    def get_description(cls):
        return ("Enable to specify the 'local-ip-prefix' when creating a "
                "security-group rule.")

    @classmethod
    def get_updated(cls):
        return "2016-03-01T10:00:00-00:00"

    def get_required_extensions(self):
        return ["security-group"]

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""NSXv add resource pool to the router bindings table

Revision ID: c288bb6a7252
Revises: b7f41687cbad
Create Date: 2016-05-15 06:12:09.450116
"""

# revision identifiers, used by Alembic.
revision = 'c288bb6a7252'
down_revision = 'b7f41687cbad'

from alembic import op
from oslo_config import cfg
import sqlalchemy as sa

from vmware_nsx.common import config  # noqa


def upgrade():
    op.add_column('nsxv_router_bindings',
                  sa.Column('resource_pool', sa.String(36), nullable=True,
                            server_default=cfg.CONF.nsxv.resource_pool_id))



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""update nsx_v3 tz_network_bindings_binding_type

Revision ID: 20483029f1ff
Revises: 69fb78b33d41
Create Date: 2016-02-09 13:57:01.590154

"""

# revision identifiers, used by Alembic.
revision = '20483029f1ff'
down_revision = '69fb78b33d41'

from alembic import op
import sqlalchemy as sa


old_tz_binding_type_enum = sa.Enum('flat', 'vlan', 'stt', 'gre', 'l3_ext',
                                   name='tz_network_bindings_binding_type')
new_tz_binding_type_enum = sa.Enum('flat', 'vlan', 'stt', 'gre', 'l3_ext',
                                   'vxlan',
                                   name='tz_network_bindings_binding_type')


def upgrade():
    op.alter_column(
        'tz_network_bindings',
        'binding_type',
        type_=new_tz_binding_type_enum,
        existing_type=old_tz_binding_type_enum,
        existing_nullable=False)



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.services.qos.notification_drivers import message_queue


class NsxV3QosNotificationDriver(
    message_queue.RpcQosServiceNotificationDriver):
    """NSXv3 message queue service notification driver for QoS.
    Overriding the create_policy method in order to add a notification
    message in this case too.
    """
    # The message queue is no longer needed in Pike.
    # Keeping this class for a while for existing configurations.
    pass



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

from neutron.db import db_base_plugin_v2
from neutron.db import l3_dvr_db  # noqa
from neutron import manager
from neutron_lib import context
from neutron_lib.plugins import constants as const
from neutron_lib.plugins import directory

from neutron_fwaas.services.firewall import fwaas_plugin as fwaas_plugin_v1
from neutron_fwaas.services.firewall import fwaas_plugin_v2

from vmware_nsx.common import config
from vmware_nsx.db import db as nsx_db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx_v3 import plugin
from vmware_nsx.plugins.nsx_v3 import utils as v3_utils
from vmware_nsx.services.fwaas.nsx_v3 import fwaas_callbacks_v1
from vmware_nsx.services.fwaas.nsx_v3 import fwaas_callbacks_v2
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsxlib.v3 import nsx_constants

_NSXLIB = None


def get_nsxv3_client(nsx_username=None, nsx_password=None,
                     use_basic_auth=False):

    return get_connected_nsxlib(nsx_username,
                                nsx_password,
                                use_basic_auth).client


def get_connected_nsxlib(nsx_username=None, nsx_password=None,
                         use_basic_auth=False):
    global _NSXLIB

    # for non-default agruments, initiate new lib
    if nsx_username or use_basic_auth:
        return v3_utils.get_nsxlib_wrapper(nsx_username,
                                           nsx_password,
                                           use_basic_auth)
    if _NSXLIB is None:
        _NSXLIB = v3_utils.get_nsxlib_wrapper()
    return _NSXLIB


def get_plugin_filters(context):
    return admin_utils.get_plugin_filters(
        context, projectpluginmap.NsxPlugins.NSX_T)


class NeutronDbClient(db_base_plugin_v2.NeutronDbPluginV2):
    def __init__(self):
        super(NeutronDbClient, self).__init__()
        self.context = context.get_admin_context()
        self.filters = get_plugin_filters(self.context)

    def _update_filters(self, requested_filters):
        filters = self.filters.copy()
        if requested_filters:
            filters.update(requested_filters)
        return filters

    def get_ports(self, filters=None, fields=None):
        filters = self._update_filters(filters)
        return super(NeutronDbClient, self).get_ports(
            self.context, filters=filters, fields=fields)

    def get_networks(self, filters=None, fields=None):
        filters = self._update_filters(filters)
        return super(NeutronDbClient, self).get_networks(
            self.context, filters=filters, fields=fields)

    def get_network(self, network_id):
        return super(NeutronDbClient, self).get_network(
            self.context, network_id)

    def get_subnet(self, subnet_id):
        return super(NeutronDbClient, self).get_subnet(self.context, subnet_id)

    def get_lswitch_and_lport_id(self, port_id):
        return nsx_db.get_nsx_switch_and_port_id(self.context.session, port_id)

    def lswitch_id_to_net_id(self, lswitch_id):
        net_ids = nsx_db.get_net_ids(self.context.session, lswitch_id)
        return net_ids[0] if net_ids else None

    def lrouter_id_to_router_id(self, lrouter_id):
        return nsx_db.get_neutron_from_nsx_router_id(self.context.session,
                                                     lrouter_id)

    def net_id_to_lswitch_id(self, net_id):
        lswitch_ids = nsx_db.get_nsx_switch_ids(self.context.session, net_id)
        return lswitch_ids[0] if lswitch_ids else None

    def add_dhcp_service_binding(self, network_id, port_id, server_id):
        return nsx_db.add_neutron_nsx_service_binding(
            self.context.session, network_id, port_id,
            nsx_constants.SERVICE_DHCP, server_id)

    def add_dhcp_static_binding(self, port_id, subnet_id, ip_address,
                                server_id, binding_id):
        return nsx_db.add_neutron_nsx_dhcp_binding(
            self.context.session, port_id, subnet_id, ip_address, server_id,
            binding_id)


class NsxV3PluginWrapper(plugin.NsxV3Plugin):
    def __init__(self):
        # initialize the availability zones
        config.register_nsxv3_azs(cfg.CONF, cfg.CONF.nsx_v3.availability_zones)
        super(NsxV3PluginWrapper, self).__init__()
        self.context = context.get_admin_context()

    def __enter__(self):
        directory.add_plugin(const.CORE, self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        directory.add_plugin(const.CORE, None)

    def _init_fwaas_plugin(self, provider, callbacks_class, plugin_callbacks):
        fwaas_plugin_class = manager.NeutronManager.load_class_for_provider(
            'neutron.service_plugins', provider)
        fwaas_plugin = fwaas_plugin_class()
        self.fwaas_callbacks = callbacks_class()
        # override the fwplugin_rpc since there is no RPC support in adminutils
        self.fwaas_callbacks.fwplugin_rpc = plugin_callbacks(fwaas_plugin)
        self.init_is_complete = True

    def init_fwaas_for_admin_utils(self):
        # initialize the FWaaS plugin and callbacks
        self.fwaas_callbacks = None
        # This is an ugly patch to find out if it is v1 or v2
        service_plugins = cfg.CONF.service_plugins
        for srv_plugin in service_plugins:
            if 'firewall' in srv_plugin:
                if 'v2' in srv_plugin:
                    # FWaaS V2
                    self._init_fwaas_plugin(
                        'firewall_v2',
                        fwaas_callbacks_v2.Nsxv3FwaasCallbacksV2,
                        fwaas_plugin_v2.FirewallCallbacks)
                else:
                    # FWaaS V1
                    self._init_fwaas_plugin(
                        'firewall',
                        fwaas_callbacks_v1.Nsxv3FwaasCallbacksV1,
                        fwaas_plugin_v1.FirewallCallbacks)
                return

    def _init_dhcp_metadata(self):
        pass

    def _process_security_group_logging(self):
        pass

    def _init_port_security_profile(self):
        return True

    def _extend_get_network_dict_provider(self, context, net):
        self._extend_network_dict_provider(context, net)
        # skip getting the Qos policy ID because get_object calls
        # plugin init again on admin-util environment

    def _extend_get_port_dict_binding(self, context, port):
        self._extend_port_dict_binding(context, port)
        # skip getting the Qos policy ID because get_object calls
        # plugin init again on admin-util environment

    def delete_network(self, network_id):
        return super(NsxV3PluginWrapper, self).delete_network(
            self.context, network_id)

    def remove_router_interface(self, router_id, interface):
        return super(NsxV3PluginWrapper, self).remove_router_interface(
            self.context, router_id, interface)



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.api.definitions import availability_zone as az_def
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import availability_zone as az_exc

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc

DEFAULT_NAME = 'default'


class ConfiguredAvailabilityZone(object):

    def __init__(self, config_line, default_name=DEFAULT_NAME):
        self.name = ""
        self._is_default = False
        if config_line and ':' in config_line:
            # Older configuration - each line contains all the relevant
            # values for one availability zones, separated by ':'
            values = config_line.split(':')
            self.name = values[0]
            self._validate_zone_name(self.name)
            self.init_from_config_line(config_line)
        elif config_line:
            # Newer configuration - the name of the availability zone can be
            # used to get the rest of the configuration for this AZ
            self.name = config_line
            self._validate_zone_name(config_line)
            self.init_from_config_section(self.name)
        else:
            # Default zone configuration
            self.name = default_name
            self._is_default = True
            self.init_default_az()

    def is_default(self):
        return self._is_default

    def _validate_zone_name(self, config_line):
        if len(self.name) > 36:
            raise nsx_exc.NsxInvalidConfiguration(
                opt_name="availability_zones",
                opt_value=config_line,
                reason=_("Maximum name length is 36"))

    @abc.abstractmethod
    def init_from_config_line(self, config_values):
        pass

    @abc.abstractmethod
    def init_from_config_section(self, az_name):
        pass

    @abc.abstractmethod
    def init_default_az(self):
        pass


class ConfiguredAvailabilityZones(object):

    default_name = DEFAULT_NAME

    def __init__(self, az_conf, az_class, default_availability_zones=None):
        self.availability_zones = {}

        # Add the configured availability zones
        for az in az_conf:
            obj = az_class(az)
            self.availability_zones[obj.name] = obj

        # add a default entry
        obj = az_class(None, default_name=self.default_name)
        self.availability_zones[obj.name] = obj

        # validate the default az:
        if default_availability_zones:
            # we support only 1 default az
            if len(default_availability_zones) > 1:
                raise nsx_exc.NsxInvalidConfiguration(
                    opt_name="default_availability_zones",
                    opt_value=default_availability_zones,
                    reason=_("The NSX plugin supports only 1 default AZ"))
            default_az_name = default_availability_zones[0]
            if (default_az_name not in self.availability_zones):
                raise nsx_exc.NsxInvalidConfiguration(
                    opt_name="default_availability_zones",
                    opt_value=default_availability_zones,
                    reason=_("The default AZ is not defined in the NSX "
                             "plugin"))
            else:
                self._default_az = self.availability_zones[default_az_name]
        else:
            self._default_az = self.availability_zones[self.default_name]

    def get_availability_zone(self, name):
        """Return an availability zone object by its name
        """
        if name in self.availability_zones.keys():
            return self.availability_zones[name]
        return self.get_default_availability_zone()

    def get_default_availability_zone(self):
        """Return the default availability zone object
        """
        return self._default_az

    def list_availability_zones(self):
        """Return a list of availability zones names
        """
        return self.availability_zones.keys()

    def list_availability_zones_objects(self):
        """Return a list of availability zones objects
        """
        return self.availability_zones.values()


class NSXAvailabilityZonesPluginCommon(object):

    @abc.abstractmethod
    def init_availability_zones(self):
        # To be initialized by the real plugin
        self._availability_zones_data = None

    def get_azs_list(self):
        return self._availability_zones_data.list_availability_zones_objects()

    def get_azs_names(self):
        return self._availability_zones_data.list_availability_zones()

    def validate_obj_azs(self, availability_zones):
        """Verify that the availability zones exist, and only 1 hint
        was set.
        """
        # For now we support only 1 hint per network/router
        # TODO(asarfaty): support multiple hints
        if len(availability_zones) > 1:
            err_msg = _("Can't use multiple availability zone hints")
            raise n_exc.InvalidInput(error_message=err_msg)

        # check that all hints appear in the predefined list of availability
        # zones
        diff = (set(availability_zones) - set(self.get_azs_names()))
        if diff:
            raise az_exc.AvailabilityZoneNotFound(
                availability_zone=diff.pop())

    def get_az_by_hint(self, hint):
        az = self._availability_zones_data.get_availability_zone(hint)
        if not az:
            raise az_def.AvailabilityZoneNotFound(availability_zone=hint)
        return az

    def get_default_az(self):
        return self._availability_zones_data.get_default_availability_zone()

    def get_obj_az_by_hints(self, obj):
        if az_def.AZ_HINTS in obj:
            for hint in obj[az_def.AZ_HINTS]:
                # For now we use only the first hint
                return self.get_az_by_hint(hint)

        # return the default
        return self.get_default_az()

    def get_network_az(self, network):
        return self.get_obj_az_by_hints(network)

    def get_router_az(self, router):
        return self.get_obj_az_by_hints(router)



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect

from vmware_nsx._i18n import _
from vmware_nsx.api_client import exception

DEFAULT_VERSION = -1


def versioned(func_table):

    def versioned_function(wrapped_func):
        func_name = wrapped_func.__name__

        def dispatch_versioned_function(cluster, *args, **kwargs):
            # Call the wrapper function, in case we need to
            # run validation checks regarding versions. It
            # should return the NSX version
            v = (wrapped_func(cluster, *args, **kwargs) or
                 cluster.api_client.get_version())
            func = get_function_by_version(func_table, func_name, v)
            func_kwargs = kwargs
            arg_spec = inspect.getargspec(func)
            if not arg_spec.keywords and not arg_spec.varargs:
                # drop args unknown to function from func_args
                arg_set = set(func_kwargs.keys())
                for arg in arg_set - set(arg_spec.args):
                    del func_kwargs[arg]
            # NOTE(salvatore-orlando): shall we fail here if a required
            # argument is not passed, or let the called function raise?
            return func(cluster, *args, **func_kwargs)

        return dispatch_versioned_function
    return versioned_function


def get_function_by_version(func_table, func_name, ver):
    if ver:
        if ver.major not in func_table[func_name]:
            major = max(func_table[func_name].keys())
            minor = max(func_table[func_name][major].keys())
            if major > ver.major:
                raise NotImplementedError(_("Operation may not be supported"))
        else:
            major = ver.major
            minor = ver.minor
            if ver.minor not in func_table[func_name][major]:
                minor = DEFAULT_VERSION
        return func_table[func_name][major][minor]
    else:
        msg = _('NSX version is not set. Unable to complete request '
                'correctly. Check log for NSX communication errors.')
        raise exception.ServiceUnavailable(message=msg)



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import hmac

import eventlet
import netaddr
from neutron_lib import constants
from neutron_lib import context as neutron_context
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log as logging

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsxv_exc
from vmware_nsx.common import locking
from vmware_nsx.common import nsxv_constants
from vmware_nsx.common import utils
from vmware_nsx.db import nsxv_db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx_v.vshield import (
    nsxv_loadbalancer as nsxv_lb)
from vmware_nsx.plugins.nsx_v.vshield.common import (
    constants as vcns_const)
from vmware_nsx.plugins.nsx_v.vshield.common import exceptions
from vmware_nsx.plugins.nsx_v.vshield import edge_utils
from vmware_nsx.services.lbaas.nsx_v import lbaas_common

METADATA_POOL_NAME = 'MDSrvPool'
METADATA_VSE_NAME = 'MdSrv'
METADATA_IP_ADDR = '169.254.169.254'
METADATA_TCP_PORT = 80
METADATA_HTTPS_PORT = 443
METADATA_HTTPS_VIP_PORT = 8775
INTERNAL_SUBNET = '169.254.128.0/17'
MAX_INIT_THREADS = 3

NET_WAIT_INTERVAL = 240
NET_CHECK_INTERVAL = 10
EDGE_WAIT_INTERVAL = 900
EDGE_CHECK_INTERVAL = 10

LOG = logging.getLogger(__name__)

DEFAULT_EDGE_FIREWALL_RULE = {
    'name': 'VSERule',
    'enabled': True,
    'action': 'allow',
    'source_vnic_groups': ['vse']}


def get_router_fw_rules():
    # build the allowed destination ports list
    int_ports = [METADATA_TCP_PORT,
                 METADATA_HTTPS_PORT,
                 METADATA_HTTPS_VIP_PORT]
    str_ports = [str(p) for p in int_ports]
    # the list of ports can be extended by configuration
    if cfg.CONF.nsxv.metadata_service_allowed_ports:
        str_metadata_ports = [str(p) for p in
                              cfg.CONF.nsxv.metadata_service_allowed_ports]
        str_ports = str_ports + str_metadata_ports
    separator = ','
    dest_ports = separator.join(str_ports)

    fw_rules = [
        DEFAULT_EDGE_FIREWALL_RULE,
        {
            'name': 'MDServiceIP',
            'enabled': True,
            'action': 'allow',
            'destination_ip_address': [METADATA_IP_ADDR],
            'protocol': 'tcp',
            'destination_port': dest_ports
        },
        {
            'name': 'MDInterEdgeNet',
            'enabled': True,
            'action': 'deny',
            'destination_ip_address': [INTERNAL_SUBNET]
        }]

    return fw_rules


def get_db_internal_edge_ips(context, az_name):
    ip_list = []
    edge_list = nsxv_db.get_nsxv_internal_edges_by_purpose(
        context.session,
        vcns_const.InternalEdgePurposes.INTER_EDGE_PURPOSE)

    if edge_list:
        # Take only the edges on this availability zone
        ip_list = [edge['ext_ip_address'] for edge in edge_list
        if nsxv_db.get_router_availability_zone(
            context.session, edge['router_id']) == az_name]
    return ip_list


class NsxVMetadataProxyHandler(object):
    """A metadata proxy handler for a specific availability zone"""
    def __init__(self, nsxv_plugin, availability_zone):
        self.nsxv_plugin = nsxv_plugin
        context = neutron_context.get_admin_context()
        self.az = availability_zone

        # Init cannot run concurrently on multiple nodes
        with locking.LockManager.get_lock('nsx-metadata-init'):
            # if the core plugin is the TVD - we need to add project
            # plugin mapping for the internal project
            core_plugin = directory.get_plugin()
            if core_plugin.is_tvd_plugin():
                try:
                    core_plugin.create_project_plugin_map(
                        context,
                        {'project_plugin_map':
                            {'plugin': projectpluginmap.NsxPlugins.NSX_V,
                             'project': nsxv_constants.INTERNAL_TENANT_ID}},
                        internal=True)
                except projectpluginmap.ProjectPluginAlreadyExists:
                    pass
            self.internal_net, self.internal_subnet = (
                self._get_internal_network_and_subnet(context))

            self.proxy_edge_ips = self._get_proxy_edges(context)

    def _create_metadata_internal_network(self, context, cidr):
        # Neutron requires a network to have some tenant_id
        tenant_id = nsxv_constants.INTERNAL_TENANT_ID
        net_name = 'inter-edge-net'
        if not self.az.is_default():
            net_name = '%s-%s' % (net_name, self.az.name)
        net_data = {'network': {'name': net_name,
                                'admin_state_up': True,
                                'port_security_enabled': False,
                                'shared': False,
                                'availability_zone_hints': [self.az.name],
                                'tenant_id': tenant_id}}
        net = self.nsxv_plugin.create_network(context, net_data)

        subnet_data = {'subnet':
                       {'cidr': cidr,
                        'name': 'inter-edge-subnet',
                        'gateway_ip': constants.ATTR_NOT_SPECIFIED,
                        'allocation_pools': constants.ATTR_NOT_SPECIFIED,
                        'ip_version': 4,
                        'dns_nameservers': constants.ATTR_NOT_SPECIFIED,
                        'host_routes': constants.ATTR_NOT_SPECIFIED,
                        'enable_dhcp': False,
                        'network_id': net['id'],
                        'tenant_id': tenant_id}}

        subnet = self.nsxv_plugin.create_subnet(
            context,
            subnet_data)

        return net['id'], subnet['id']

    def _get_internal_net_by_az(self, context):
        # Get the internal network for the current az
        int_net = nsxv_db.get_nsxv_internal_network_for_az(
            context.session,
            vcns_const.InternalEdgePurposes.INTER_EDGE_PURPOSE,
            self.az.name)

        if int_net:
            return int_net['network_id']

    def _get_internal_network_and_subnet(self, context):
        # Try to find internal net, internal subnet. If not found, create new
        internal_net = self._get_internal_net_by_az(context)
        internal_subnet = None

        if internal_net:
            internal_subnet = self.nsxv_plugin.get_subnets(
                context,
                fields=['id'],
                filters={'network_id': [internal_net]})[0]['id']

        if internal_net is None or internal_subnet is None:
            if cfg.CONF.nsxv.metadata_initializer:
                # Couldn't find net, subnet - create new
                try:
                    internal_net, internal_subnet = (
                        self._create_metadata_internal_network(
                            context, INTERNAL_SUBNET))
                except Exception as e:
                    nsxv_db.delete_nsxv_internal_network(
                        context.session,
                        vcns_const.InternalEdgePurposes.INTER_EDGE_PURPOSE,
                        internal_net)

                    # if network is created, clean up
                    if internal_net:
                        self.nsxv_plugin.delete_network(context,
                                                        internal_net)

                    error = (_("Exception %s while creating internal "
                               "network for metadata service") % e)
                    LOG.exception(error)
                    raise nsxv_exc.NsxPluginException(err_msg=error)

                # Update the new network_id in DB
                nsxv_db.create_nsxv_internal_network(
                    context.session,
                    nsxv_constants.INTER_EDGE_PURPOSE,
                    self.az.name,
                    internal_net)
            else:
                error = _('Metadata initialization is incomplete on '
                          'initializer node')
                raise nsxv_exc.NsxPluginException(err_msg=error)

        return internal_net, internal_subnet

    def _get_edge_internal_ip(self, context, rtr_id):
        filters = {
            'network_id': [self.internal_net],
            'device_id': [rtr_id]}
        ports = self.nsxv_plugin.get_ports(context, filters=filters)
        if ports:
            return ports[0]['fixed_ips'][0]['ip_address']
        else:
            LOG.error("No port found for metadata for %s", rtr_id)

    def _get_edge_rtr_id_by_ext_ip(self, context, edge_ip):
        rtr_list = nsxv_db.get_nsxv_internal_edge(
            context.session, edge_ip)
        if rtr_list:
            return rtr_list[0]['router_id']

    def _get_edge_id_by_rtr_id(self, context, rtr_id):
        binding = nsxv_db.get_nsxv_router_binding(
            context.session,
            rtr_id)

        if binding:
            return binding['edge_id']

    def _get_proxy_edges(self, context):
        proxy_edge_ips = []

        db_edge_ips = get_db_internal_edge_ips(context, self.az.name)
        if len(db_edge_ips) > len(self.az.mgt_net_proxy_ips):
            error = (_('Number of configured metadata proxy IPs is smaller '
                      'than number of Edges which are already provisioned '
                      'for availability zone %s'), self.az.name)
            raise nsxv_exc.NsxPluginException(err_msg=error)

        pool = eventlet.GreenPool(min(MAX_INIT_THREADS,
                                      len(self.az.mgt_net_proxy_ips)))

        # Edge IPs that exist in both lists have to be validated that their
        # Edge appliance settings are valid
        for edge_inner_ip in pool.imap(
                self._setup_proxy_edge_route_and_connectivity,
                list(set(db_edge_ips) & set(self.az.mgt_net_proxy_ips))):
            proxy_edge_ips.append(edge_inner_ip)

        # Edges that exist only in the CFG list, should be paired with Edges
        # that exist only in the DB list. The existing Edge from the list will
        # be reconfigured to match the new config
        edge_to_convert_ips = (
            list(set(db_edge_ips) - set(self.az.mgt_net_proxy_ips)))
        edge_ip_to_set = (
            list(set(self.az.mgt_net_proxy_ips) - set(db_edge_ips)))

        if edge_to_convert_ips:
            if cfg.CONF.nsxv.metadata_initializer:
                for edge_inner_ip in pool.imap(
                        self._setup_proxy_edge_external_interface_ip,
                        zip(edge_to_convert_ips, edge_ip_to_set)):
                    proxy_edge_ips.append(edge_inner_ip)
            else:
                error = _('Metadata initialization is incomplete on '
                          'initializer node')
                raise nsxv_exc.NsxPluginException(err_msg=error)

        # Edges that exist in the CFG list but do not have a matching DB
        # element will be created.
        remaining_cfg_ips = edge_ip_to_set[len(edge_to_convert_ips):]
        if remaining_cfg_ips:
            if cfg.CONF.nsxv.metadata_initializer:
                for edge_inner_ip in pool.imap(
                        self._setup_new_proxy_edge, remaining_cfg_ips):
                    proxy_edge_ips.append(edge_inner_ip)

                pool.waitall()
            else:
                error = _('Metadata initialization is incomplete on '
                          'initializer node')
                raise nsxv_exc.NsxPluginException(err_msg=error)

        return proxy_edge_ips

    def _setup_proxy_edge_route_and_connectivity(self, rtr_ext_ip,
                                                 rtr_id=None, edge_id=None):
        # Use separate context per each as we use this in tread context
        context = neutron_context.get_admin_context()
        if not rtr_id:
            rtr_id = self._get_edge_rtr_id_by_ext_ip(context, rtr_ext_ip)
        if not edge_id:
            edge_id = self._get_edge_id_by_rtr_id(context, rtr_id)
        if not rtr_id or not edge_id:
            # log this error and return without the ip, but don't fail
            LOG.error("Failed find edge for router %(rtr_id)s with ip "
                      "%(rtr_ext_ip)s",
                      {'rtr_id': rtr_id, 'rtr_ext_ip': rtr_ext_ip})
            return

        # Read and validate DGW. If different, replace with new value
        try:
            # This may fail if the edge was deleted on backend
            h, routes = self.nsxv_plugin.nsx_v.vcns.get_routes(edge_id)
        except exceptions.ResourceNotFound as e:
            # log this error and return without the ip, but don't fail
            LOG.error("Failed to get routes for metadata proxy edge "
                      "%(edge)s: %(err)s",
                      {'edge': edge_id, 'err': e})
            return

        dgw = routes.get('defaultRoute', {}).get('gatewayAddress')

        if dgw != self.az.mgt_net_default_gateway:
            if cfg.CONF.nsxv.metadata_initializer:
                self.nsxv_plugin._update_routes(
                    context, rtr_id,
                    self.az.mgt_net_default_gateway)
            else:
                error = _('Metadata initialization is incomplete on '
                          'initializer node')
                raise nsxv_exc.NsxPluginException(err_msg=error)

        # Read and validate connectivity
        h, if_data = self.nsxv_plugin.nsx_v.get_interface(
            edge_id, vcns_const.EXTERNAL_VNIC_INDEX)
        cur_ip = if_data.get('addressGroups', {}
                             ).get('addressGroups', {}
                                   )[0].get('primaryAddress')
        cur_pgroup = if_data['portgroupId']
        if (if_data and cur_pgroup != self.az.mgt_net_moid
                or cur_ip != rtr_ext_ip):
            if cfg.CONF.nsxv.metadata_initializer:
                self.nsxv_plugin.nsx_v.update_interface(
                    rtr_id,
                    edge_id,
                    vcns_const.EXTERNAL_VNIC_INDEX,
                    self.az.mgt_net_moid,
                    address=rtr_ext_ip,
                    netmask=self.az.mgt_net_proxy_netmask,
                    secondary=[])
            else:
                error = _('Metadata initialization is incomplete on '
                          'initializer node')
                raise nsxv_exc.NsxPluginException(err_msg=error)

        # Read and validate LB pool member configuration
        # When the Nova IP address is changed in the ini file, we should apply
        # this change to the LB pool
        lb_obj = nsxv_lb.NsxvLoadbalancer.get_loadbalancer(
            self.nsxv_plugin.nsx_v.vcns, edge_id)

        vs = lb_obj.virtual_servers.get(METADATA_VSE_NAME)
        update_md_proxy = False
        if vs:
            md_members = {member.payload['ipAddress']: member.payload['name']
                          for member in vs.default_pool.members.values()}

            if len(cfg.CONF.nsxv.nova_metadata_ips) == len(md_members):
                m_ips = md_members.keys()
                m_to_convert = (list(set(m_ips) -
                                     set(cfg.CONF.nsxv.nova_metadata_ips)))
                m_ip_to_set = (list(set(cfg.CONF.nsxv.nova_metadata_ips)
                                    - set(m_ips)))
                if m_to_convert or m_ip_to_set:
                    update_md_proxy = True
                for m_ip in m_to_convert:
                    m_name = md_members[m_ip]
                    vs.default_pool.members[m_name].payload['ipAddress'] = (
                        m_ip_to_set.pop())
            else:
                error = _('Number of metadata members should not change')
                raise nsxv_exc.NsxPluginException(err_msg=error)

            try:
                # This may fail if the edge is powered off right now
                if update_md_proxy:
                    lb_obj.submit_to_backend(self.nsxv_plugin.nsx_v.vcns,
                                             edge_id)
            except exceptions.RequestBad as e:
                # log the error and continue
                LOG.error("Failed to update load balancer on metadata "
                          "proxy edge %(edge)s: %(err)s",
                          {'edge': edge_id, 'err': e})

        edge_ip = self._get_edge_internal_ip(context, rtr_id)

        if edge_ip:
            return edge_ip

    def _setup_proxy_edge_external_interface_ip(self, rtr_ext_ips):
        # Use separate context per each as we use this in tread context
        context = neutron_context.get_admin_context()

        rtr_old_ext_ip, rtr_new_ext_ip = rtr_ext_ips

        rtr_id = self._get_edge_rtr_id_by_ext_ip(context, rtr_old_ext_ip)
        edge_id = self._get_edge_id_by_rtr_id(context, rtr_id)

        # Replace DB entry as we cannot update the table PK
        nsxv_db.delete_nsxv_internal_edge(context.session, rtr_old_ext_ip)

        edge_ip = self._setup_proxy_edge_route_and_connectivity(
            rtr_new_ext_ip, rtr_id, edge_id)

        nsxv_db.create_nsxv_internal_edge(
            context.session, rtr_new_ext_ip,
            vcns_const.InternalEdgePurposes.INTER_EDGE_PURPOSE, rtr_id)

        if edge_ip:
            return edge_ip

    def _setup_new_proxy_edge(self, rtr_ext_ip):
        # Use separate context per each as we use this in tread context
        context = neutron_context.get_admin_context()

        rtr_id = None

        try:
            rtr_name = 'metadata_proxy_router'
            if not self.az.is_default():
                rtr_name = '%s-%s' % (rtr_name, self.az.name)
            router_data = {
                'router': {
                    'name': rtr_name,
                    'admin_state_up': True,
                    'router_type': 'exclusive',
                    'availability_zone_hints': [self.az.name],
                    'tenant_id': nsxv_constants.INTERNAL_TENANT_ID}}

            rtr = self.nsxv_plugin.create_router(
                context,
                router_data,
                allow_metadata=False)

            rtr_id = rtr['id']
            edge_id = self._get_edge_id_by_rtr_id(context, rtr_id)
            if not edge_id:
                LOG.error('No edge create for router - %s', rtr_id)
                if rtr_id:
                    self.nsxv_plugin.delete_router(context, rtr_id)
                return

            self.nsxv_plugin.nsx_v.update_interface(
                rtr['id'],
                edge_id,
                vcns_const.EXTERNAL_VNIC_INDEX,
                self.az.mgt_net_moid,
                address=rtr_ext_ip,
                netmask=self.az.mgt_net_proxy_netmask,
                secondary=[])

            port_data = {
                'port': {
                    'network_id': self.internal_net,
                    'name': None,
                    'admin_state_up': True,
                    'device_id': rtr_id,
                    'device_owner': (constants.DEVICE_OWNER_NETWORK_PREFIX +
                                     'md_interface'),
                    'fixed_ips': constants.ATTR_NOT_SPECIFIED,
                    'mac_address': constants.ATTR_NOT_SPECIFIED,
                    'port_security_enabled': False,
                    'tenant_id': nsxv_constants.INTERNAL_TENANT_ID}}

            port = self.nsxv_plugin.base_create_port(context, port_data)

            address_groups = self._get_address_groups(
                context, self.internal_net, rtr_id, is_proxy=True)

            edge_ip = port['fixed_ips'][0]['ip_address']
            with locking.LockManager.get_lock(edge_id):
                edge_utils.update_internal_interface(
                    self.nsxv_plugin.nsx_v, context, rtr_id,
                    self.internal_net, address_groups)

            self._setup_metadata_lb(rtr_id,
                                    port['fixed_ips'][0]['ip_address'],
                                    cfg.CONF.nsxv.nova_metadata_port,
                                    cfg.CONF.nsxv.nova_metadata_port,
                                    cfg.CONF.nsxv.nova_metadata_ips,
                                    proxy_lb=True)

            firewall_rules = [
                DEFAULT_EDGE_FIREWALL_RULE,
                {
                    'action': 'allow',
                    'enabled': True,
                    'source_ip_address': [INTERNAL_SUBNET]}]

            edge_utils.update_firewall(
                self.nsxv_plugin.nsx_v,
                context,
                rtr_id,
                {'firewall_rule_list': firewall_rules},
                allow_external=False)

            if self.az.mgt_net_default_gateway:
                self.nsxv_plugin._update_routes(
                    context, rtr_id,
                    self.az.mgt_net_default_gateway)

            nsxv_db.create_nsxv_internal_edge(
                context.session, rtr_ext_ip,
                vcns_const.InternalEdgePurposes.INTER_EDGE_PURPOSE, rtr_id)

            return edge_ip

        except Exception as e:
            LOG.exception("Exception %s while creating internal edge "
                          "for metadata service", e)

            ports = self.nsxv_plugin.get_ports(
                context, filters={'device_id': [rtr_id]})

            for port in ports:
                self.nsxv_plugin.delete_port(context, port['id'],
                                             l3_port_check=True,
                                             nw_gw_port_check=True,
                                             allow_delete_internal=True)

            nsxv_db.delete_nsxv_internal_edge(
                context.session,
                rtr_ext_ip)

            if rtr_id:
                self.nsxv_plugin.delete_router(context, rtr_id)

    def _get_address_groups(self, context, network_id, device_id, is_proxy):

        filters = {'network_id': [network_id],
                   'device_id': [device_id]}
        ports = self.nsxv_plugin.get_ports(context, filters=filters)

        subnets = self.nsxv_plugin.get_subnets(context, filters=filters)

        address_groups = []
        for subnet in subnets:
            address_group = {}
            net = netaddr.IPNetwork(subnet['cidr'])
            address_group['subnetMask'] = str(net.netmask)
            address_group['subnetPrefixLength'] = str(net.prefixlen)
            for port in ports:
                fixed_ips = port['fixed_ips']
                for fip in fixed_ips:
                    s_id = fip['subnet_id']
                    ip_addr = fip['ip_address']
                    if s_id == subnet['id'] and netaddr.valid_ipv4(ip_addr):
                        address_group['primaryAddress'] = ip_addr
                        break

            # For Edge appliances which aren't the metadata proxy Edge
            #  we add the metadata IP address
            if not is_proxy and network_id == self.internal_net:
                address_group['secondaryAddresses'] = {
                    'type': 'secondary_addresses',
                    'ipAddress': [METADATA_IP_ADDR]}

            address_groups.append(address_group)
        return address_groups

    def _create_ssl_cert(self, edge_id=None):
        # Create a self signed certificate in the backend if both Cert details
        # and private key are not supplied in nsx.ini
        if (not cfg.CONF.nsxv.metadata_nova_client_cert and
            not cfg.CONF.nsxv.metadata_nova_client_priv_key):
            h = self.nsxv_plugin.nsx_v.vcns.create_csr(edge_id)[0]
            # Extract the CSR ID from header
            csr_id = lbaas_common.extract_resource_id(h['location'])
            # Create a self signed certificate
            cert = self.nsxv_plugin.nsx_v.vcns.create_csr_cert(csr_id)[1]
            cert_id = cert['objectId']
        else:
            # Raise an error if either the Cert path or the private key is not
            # configured
            error = None
            if not cfg.CONF.nsxv.metadata_nova_client_cert:
                error = _('Metadata certificate path not configured')
            elif not cfg.CONF.nsxv.metadata_nova_client_priv_key:
                error = _('Metadata client private key not configured')
            if error:
                raise nsxv_exc.NsxPluginException(err_msg=error)
            pem_encoding = utils.read_file(
                cfg.CONF.nsxv.metadata_nova_client_cert)
            priv_key = utils.read_file(
                cfg.CONF.nsxv.metadata_nova_client_priv_key)
            request = {
                'pemEncoding': pem_encoding,
                'privateKey': priv_key}
            cert = self.nsxv_plugin.nsx_v.vcns.upload_edge_certificate(
                edge_id, request)[1]
            cert_id = cert.get('certificates')[0]['objectId']
        return cert_id

    def _setup_metadata_lb(self, rtr_id, vip, v_port, s_port, member_ips,
                           proxy_lb=False, context=None):

        if context is None:
            context = neutron_context.get_admin_context()

        edge_id = self._get_edge_id_by_rtr_id(context, rtr_id)
        LOG.debug('Setting up Edge device %s', edge_id)

        lb_obj = nsxv_lb.NsxvLoadbalancer()

        protocol = 'HTTP'
        ssl_pass_through = False
        cert_id = None
        # Set protocol to HTTPS with default port of 443 if metadata_insecure
        # is set to False.
        if not cfg.CONF.nsxv.metadata_insecure:
            protocol = 'HTTPS'
            if proxy_lb:
                v_port = METADATA_HTTPS_VIP_PORT
            else:
                v_port = METADATA_HTTPS_PORT
                # Create the certificate on the backend
                cert_id = self._create_ssl_cert(edge_id)
            ssl_pass_through = proxy_lb
        mon_type = protocol if proxy_lb else 'tcp'
        # Create virtual server
        virt_srvr = nsxv_lb.NsxvLBVirtualServer(
            name=METADATA_VSE_NAME,
            ip_address=vip,
            protocol=protocol,
            port=v_port)

        # For router Edge, we add X-LB-Proxy-ID header
        if not proxy_lb:
            md_app_rule = nsxv_lb.NsxvLBAppRule(
                'insert-mdp',
                'reqadd X-Metadata-Provider:' + edge_id)
            virt_srvr.add_app_rule(md_app_rule)

            # When shared proxy is configured, insert authentication string
            if cfg.CONF.nsxv.metadata_shared_secret:
                signature = hmac.new(
                    cfg.CONF.nsxv.metadata_shared_secret,
                    edge_id,
                    hashlib.sha256).hexdigest()
                sign_app_rule = nsxv_lb.NsxvLBAppRule(
                    'insert-auth',
                    'reqadd X-Metadata-Provider-Signature:' + signature)
                virt_srvr.add_app_rule(sign_app_rule)

        # Create app profile
        #  XFF is inserted in router LBs
        app_profile = nsxv_lb.NsxvLBAppProfile(
            name='MDSrvProxy',
            template=protocol,
            server_ssl_enabled=not cfg.CONF.nsxv.metadata_insecure,
            ssl_pass_through=ssl_pass_through,
            insert_xff=not proxy_lb,
            client_ssl_cert=cert_id)

        virt_srvr.set_app_profile(app_profile)

        # Create pool, members and monitor
        pool = nsxv_lb.NsxvLBPool(
            name=METADATA_POOL_NAME)

        monitor = nsxv_lb.NsxvLBMonitor(name='MDSrvMon',
                                        mon_type=mon_type.lower())
        pool.add_monitor(monitor)

        i = 0
        for member_ip in member_ips:
            i += 1
            member = nsxv_lb.NsxvLBPoolMember(
                name='Member-%d' % i,
                ip_address=member_ip,
                port=s_port,
                monitor_port=s_port)
            pool.add_member(member)

        virt_srvr.set_default_pool(pool)
        lb_obj.add_virtual_server(virt_srvr)

        lb_obj.submit_to_backend(self.nsxv_plugin.nsx_v.vcns, edge_id)

    def configure_router_edge(self, context, rtr_id):
        ctx = context.elevated()
        # Connect router interface to inter-edge network
        port_data = {
            'port': {
                'network_id': self.internal_net,
                'name': None,
                'admin_state_up': True,
                'device_id': rtr_id,
                'device_owner': constants.DEVICE_OWNER_ROUTER_GW,
                'fixed_ips': constants.ATTR_NOT_SPECIFIED,
                'mac_address': constants.ATTR_NOT_SPECIFIED,
                'port_security_enabled': False,
                'tenant_id': nsxv_constants.INTERNAL_TENANT_ID}}

        self.nsxv_plugin.base_create_port(ctx, port_data)

        address_groups = self._get_address_groups(
            ctx,
            self.internal_net,
            rtr_id,
            is_proxy=False)

        edge_utils.update_internal_interface(
            self.nsxv_plugin.nsx_v,
            context,
            rtr_id,
            self.internal_net,
            address_groups=address_groups)

        self._setup_metadata_lb(rtr_id,
                                METADATA_IP_ADDR,
                                METADATA_TCP_PORT,
                                cfg.CONF.nsxv.nova_metadata_port,
                                self.proxy_edge_ips,
                                proxy_lb=False,
                                context=context)

    def cleanup_router_edge(self, context, rtr_id, warn=False):
        filters = {
            'network_id': [self.internal_net],
            'device_id': [rtr_id]}
        ctx = context.elevated()
        ports = self.nsxv_plugin.get_ports(ctx, filters=filters)

        if ports:
            if warn:
                LOG.warning("cleanup_router_edge found port %(port)s for "
                            "router %(router)s - deleting it now.",
                            {'port': ports[0]['id'], 'router': rtr_id})
            try:
                self.nsxv_plugin.delete_port(
                    ctx, ports[0]['id'],
                    l3_port_check=False)
            except Exception as e:
                LOG.error("Failed to delete md_proxy port %(port)s: "
                          "%(e)s", {'port': ports[0]['id'], 'e': e})

    def is_md_subnet(self, subnet_id):
        return self.internal_subnet == subnet_id



# Copyright 2012 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import six

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions

LOG = logging.getLogger(__name__)
DEFAULT_PORT = 443
# Raise if one of those attributes is not specified
REQUIRED_ATTRIBUTES = ['default_tz_uuid', 'nsx_user',
                       'nsx_password', 'nsx_controllers']
# Emit an INFO log if one of those attributes is not specified
IMPORTANT_ATTRIBUTES = ['default_l3_gw_service_uuid']
# Deprecated attributes
DEPRECATED_ATTRIBUTES = ['metadata_dhcp_host_route',
                         'nvp_user', 'nvp_password', 'nvp_controllers']


class NSXCluster(object):
    """NSX cluster class.

    Encapsulates controller connections and the API client for a NSX cluster.

    Controller-specific parameters, such as timeouts are stored in the
    elements of the controllers attribute, which are dicts.
    """

    def __init__(self, **kwargs):
        self._required_attributes = REQUIRED_ATTRIBUTES[:]
        self._important_attributes = IMPORTANT_ATTRIBUTES[:]
        self._deprecated_attributes = {}
        self._sanity_check(kwargs)

        for opt, val in six.iteritems(self._deprecated_attributes):
            LOG.deprecated(_("Attribute '%s' has been deprecated or moved "
                             "to a new section. See new configuration file "
                             "for details."), opt)
            depr_func = getattr(self, '_process_%s' % opt, None)
            if depr_func:
                depr_func(val)

        # If everything went according to plan these two lists should be empty
        if self._required_attributes:
            raise exceptions.InvalidClusterConfiguration(
                invalid_attrs=self._required_attributes)
        if self._important_attributes:
            LOG.info("The following cluster attributes were "
                     "not specified: %s'", self._important_attributes)
        # The API client will be explicitly created by users of this class
        self.api_client = None

    def _sanity_check(self, options):
        # Iterating this way ensures the conf parameters also
        # define the structure of this class
        for arg in cfg.CONF:
            if arg not in DEPRECATED_ATTRIBUTES:
                setattr(self, arg, options.get(arg, cfg.CONF.get(arg)))
                self._process_attribute(arg)
            elif options.get(arg) is not None:
                # Process deprecated attributes only if specified
                self._deprecated_attributes[arg] = options.get(arg)

    def _process_attribute(self, attribute):
        # Process the attribute only if it's not empty!
        if getattr(self, attribute, None):
            if attribute in self._required_attributes:
                self._required_attributes.remove(attribute)
            if attribute in self._important_attributes:
                self._important_attributes.remove(attribute)
            handler_func = getattr(self, '_process_%s' % attribute, None)
            if handler_func:
                handler_func()

    def _process_nsx_controllers(self):
        # If this raises something is not right, so let it bubble up
        # TODO(salvatore-orlando): Also validate attribute here
        for i, ctrl in enumerate(self.nsx_controllers or []):
            if len(ctrl.split(':')) == 1:
                self.nsx_controllers[i] = '%s:%s' % (ctrl, DEFAULT_PORT)

    def _process_nvp_controllers(self):
        self.nsx_controllers = self.nvp_controllers
        self._process_nsx_controllers()
